/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
 *  afmsup.c -- support for AFM files
 *
 *      jrs     30.05.98
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "cfg.h"
#include "utils.h"
#include "lpudefs.h"
#include "devapi.h"
#include "lpuvars.h"

char    *Isotab[] = {
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "",
// 040
  "space",
  "exclam",
  "quotedbl",
  "numbersign",
  "dollar",
  "percent",
  "ampersand",
  "quoteright",
// 050
  "parenleft",
  "parenright",
  "asterisk",
  "plus",
  "comma",
  "hyphen",
  "period",
  "slash",
// 060
  "zero",
  "one",
  "two",
  "three",
  "four",
  "five",
  "six",
  "seven",
// 070
  "eight",
  "nine",
  "colon",
  "semicolon",
  "less",
  "equal",
  "greater",
  "question",
// 100
  "at",
  "A",
  "B",
  "C",
  "D",
  "E",
  "F",
  "G",
// 110
  "H",
  "I",
  "J",
  "K",
  "L",
  "M",
  "N",
  "O",
// 120
  "P",
  "Q",
  "R",
  "S",
  "T",
  "U",
  "V",
  "W",
// 130
  "X",
  "Y",
  "Z",
  "bracketleft",
  "backslash",
  "bracketright",
  "asciicircum",
  "underscore",
// 140
  "quoteleft",
  "a",
  "b",
  "c",
  "d",
  "e",
  "f",
  "g",
// 150
  "h",
  "i",
  "j",
  "k",
  "l",
  "m",
  "n",
  "o",
// 160
  "p",
  "q",
  "r",
  "s",
  "t",
  "u",
  "v",
  "w",
// 170
  "x",
  "y",
  "z",
  "braceleft",
  "bar",
  "braceright",
  "asciitilde",
  "",
// 200
  "", "", "", "", "", "", "", "",
// 210
  "", "", "", "", "", "", "", "",
// 220
  "dotlessi",
  "grave",
  "acute",
  "circumflex",
  "tilde",
  "",
  "breve",
  "dotaccent",
// 230
  "",  // ???
  "",
  "ring",
  "",  // ???
  "",
  "hungarumlaut",
  "ogonek",
  "caron",
// 240
  "",
  "exclamdown",
  "cent",
  "sterling",
  "currency",
  "yen",
  "brokenbar",
  "section",
// 250
  "dieresis",
  "copyright",
  "ordfeminine",
  "guillemotleft",
  "logicalnot",
  "hyphen",
  "registered",
  "macron",
// 260
  "degree",
  "plusminus",
  "twosuperior",
  "threesuperior",
  "acute",
  "mu",
  "paragraph",
  "periodcentered",
// 270
  "cedilla",
  "onesuperior",
  "ordmasculine",
  "guillemotright",
  "onequarter",
  "onehalf",
  "threequarters",
  "questiondown",
// 300
  "Agrave",
  "Aacute",
  "Acircumflex",
  "Atilde",
  "Adieresis",
  "Aring",
  "AE",
  "Ccedilla",
// 310
  "Egrave",
  "Eacute",
  "Ecircumflex",
  "Edieresis",
  "Igrave",
  "Iacute",
  "Icircumflex",
  "Idieresis",
// 320
  "Eth",
  "Ntilde",
  "Ograve",
  "Oacute",
  "Ocircumflex",
  "Otilde",
  "Odieresis",
  "multiply",
// 330
  "Oslash",
  "Ugrave",
  "Uacute",
  "Ucircumflex",
  "Udieresis",
  "Yacute",
  "Thorn",
  "germandbls",
// 340
  "agrave",
  "aacute",
  "acircumflex",
  "atilde",
  "adieresis",
  "aring",
  "ae",
  "ccedilla",
// 350
  "egrave",
  "eacute",
  "ecircumflex",
  "edieresis",
  "igrave",
  "iaccute",
  "icircumflex",
  "idieresis",
// 360
  "eth",
  "ntilde",
  "ograve",
  "oacute",
  "ocircumflex",
  "otilde",
  "odieresis",
  "divide",
// 370
  "oslash",
  "ugrave",
  "uacute",
  "ucircumflex",
  "udieresis",
  "yacute",
  "thorn",
  "ydieresis"
};


int isoname2ndx(char *name)
{
    int     i;

    for (i=32; i<256; i++)
        if (!strcmp(name, Isotab[i]))
            return i;
    return 0;
}


static int afm_fontname(tPSFONT *font)
{
    skipgap();
    strcpy(font->fontname, Cp);
    return TRUE;
}

static int afm_isfixed(tPSFONT *font)
{
    char    buf[10];

    skipgap();
    if (!getname(buf, 9, TRUE))
        return dev_error(devErrBAD, "parameter expected");
    if (!strcmp(buf, "FALSE"))
    {
        font->p_flag   = FALSE;
        font->relwidth = 600;
        return TRUE;
    }
    if (!strcmp(buf, "TRUE"))
    {
        font->p_flag   = TRUE;
        return TRUE;
    }
    return dev_error(devErrBAD, "boolean expected");
}

static int afm_char(tPSFONT *font)
{
    int     c, wx, n;
    char    buf[65];

    skipgap();
    if (!getdec(&c))
        return dev_error(devErrBAD, "C-value expected");
    skipgap();
    if (!needsemi())
        return dev_error(devErrBAD, "semicolon expected");
    skipgap();
    if (!getname(buf, 9, FALSE) || strcmp(buf, "WX"))
        return dev_error(devErrBAD, "WX expected");
    skipgap();
    if (!getdec(&wx))
        return dev_error(devErrBAD, "WX-value expected");
    skipgap();
    if (!needsemi())
        return dev_error(devErrBAD, "semicolon expected");
    skipgap();
    if (!getname(buf, 9, FALSE) || strcmp(buf, "N"))
        return dev_error(devErrBAD, "N expected");
    skipgap();
    if (*Cp == '.')
        return TRUE;
    if (!getname(buf, 64, FALSE))
        return dev_error(devErrBAD, "charname expected");
    n = isoname2ndx(buf);
    if (n == 0)
        return TRUE;
    font->widtab[n] = wx;
    return TRUE;
}

static int kpxcnt = 0;
static int kpxndx = 0;

static int afm_kpx_init(tPSFONT *font)
{
    tKPX    *p;

    skipgap();
    if (!getdec(&kpxcnt))
        return dev_error(devErrBAD, "kernpair-count expected");
    p = (tKPX *) calloc(sizeof(tKPX), kpxcnt);
    if (p == NULL)
        return dev_error(devErrBAD, "out of memory");
    font->kerntab = p;
    font->kerncnt = 0;
    kpxndx = 0;
    return 1; 
}

static int afm_kpx(tPSFONT *font)
{
    tKPX    *p;
    char    buf[65];
    int     n;

    if (font->kerntab == NULL)
        return error("unexpected KPX line");
    if (kpxndx >= kpxcnt)
        return error("too many KPX lines");
    p = font->kerntab + kpxndx;
    skipgap();
    if (!getname(buf, 64, FALSE))
        return error("charname expected");
    p->first = isoname2ndx(buf);
    if (p->first == 0)
        return 1;
    skipgap();
    if (!getname(buf, 64, FALSE))
        return error("charname expected");
    p->second = isoname2ndx(buf);
    if (p->second == 0)
        return 1;
    skipgap();
    if (!getdec(&n))
        return error("value expected");
    p->shift = n;
    kpxndx++;
    font->kerncnt = kpxndx;
    return 1;
}

static int afm_upos(tPSFONT *font)
{
    skipgap();
    if (!getdec(&(font->upos)))
        return error("underline position expected");
    return 1;
}

static int afm_uwid(tPSFONT *font)
{
    skipgap();
    if (!getdec(&(font->uwid)))
        return error("underline thickness expected");
    return 1;
}

static int afm_fontbbox(tPSFONT *font)
{
    skipgap();
    if (!getdec(&(font->bbx0)))
        return error("bounding box x0 expected");
    skipgap();
    if (!getdec(&(font->bby0)))
        return error("bounding box y0 expected");
    skipgap();
    if (!getdec(&(font->bbx1)))
        return error("bounding box x1 expected");
    skipgap();
    if (!getdec(&(font->bby1)))
        return error("bounding box y1 expected");
    return 1;
}

static int do_afmcmd(tPSFONT *font, char *cmd)
{
    if (!strcmp(cmd, "FontName"))           return afm_fontname(font);
    if (!strcmp(cmd, "FontBBox"))           return afm_fontbbox(font);
    if (!strcmp(cmd, "IsFixedPitch"))       return afm_isfixed(font);
    if (!strcmp(cmd, "UnderlinePosition"))  return afm_upos(font);
    if (!strcmp(cmd, "UnderlineThickness")) return afm_uwid(font);
    if (!strcmp(cmd, "StartKernPairs"))     return afm_kpx_init(font);
    if (!strcmp(cmd, "C"))                  return afm_char(font);
    if (!strcmp(cmd, "KPX"))                return afm_kpx(font);
    return TRUE;
}

int scan_afmfile(tPSFONT *font, char *filename)
{
    int     n;
    char    fn[128], buf[256], cmd[64];
    FILE    *fp;

    sprintf(fn, "%s/%s.afm", Afmdir, filename);
    fp = fopen(fn, "r");
    if (fp == NULL)
        return dev_error(devErrBAD, "unknown filename '%s'", fn);
    font->kerncnt = 0;
    font->kerntab = NULL;
    kpxndx = 0;
    while (!feof(fp))
    {
        fgets(buf, 255, fp);
        if (feof(fp))
            continue;
        Cp = buf;
        if (!getname(cmd, 63, FALSE))
            continue;
        if (!do_afmcmd(font, cmd))
            return FALSE;
    }
    fclose(fp);
    if (font->p_flag)
    {
        n = isoname2ndx("zero");
        if (n >= 0)
            font->relwidth = font->widtab[n];
    }
    return 1;
}

