/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
 *  devapi.c -- Hilfsfunktionen fuer DEV-Treiber
 *
 *      jrs     04.09.97
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include "cfg.h"
#include "utils.h"
#include "lpudefs.h"
#include "devapi.h"
#include "afmsup.h"
#include "lpuvars.h"

static  tDEV    Dev;
static  tPAGE   *pPag;
static  tPAGE   Pages[MAXPAGES];
static  int     Pagcnt;
static  int     Pagnr;

static  tPSFONT *pFnt;
static  tPSFONT Fonts[MAXFONTS];
static  tFONT   Myfont;
static  int     Fntcnt;
static  int     Fntref = 0;
static  int     Dashndx = 0;

static  FILE    *Ofp;
static  int     Xadj,  Yadj;    /* Papier-Justage                           */
static  int     Xbase, Ybase;   /* linke obere Ecke des aktuellen Fensters  */
static  int     Xsiz,  Ysiz;    /* Groesse des aktuellen Ausgabefensters    */
static  int     Xpos,  Ypos;    /* rel. Schreibposition im Fenster          */
static	int     Xmax,  Ymax;
static  int     Lm, Rm;         /* kleinste und groesste nutzbare X-Koord.  */
static  int     Tm, Bm;         /* kleinste und groesste nutzbare Y-Koord.  */
static  int     Rotate;         /* rotation angle for the current page      */
static  int     Xshift, Yshift; /* shifting offsets for the current page    */
static  int     Xscale, Yscale; /* scaling factors for the current page     */


typedef struct
{
    int     c;
    int     m;
    int     y;
    int     k;
} CMYK;

static  int     CurrentPen  = -1;
static  int     SelectedPen = 10;
static  tCMYK   DefaultInk  = {   0,   0,   0, 100 };
static  tCMYK   CurrentInk  = {  -1,  -1,  -1,  -1 };
static  tCMYK   SelectedInk = {   0,   0,   0, 100 };

int dev_error(int rc, char *fmt, ...)
{
    va_list arglst;
    char    buf[256];

    va_start(arglst, fmt);
    vsprintf(buf, fmt, arglst);
    va_end(arglst);
    fprintf(stderr,"error: %s\n",buf);
    return rc;
}

static tPAGE *findpage(char *id)
{
    int     i;

    i = 0;
    while (i < Pagcnt)
    {
        if (!strcmp(id, Pages[i].id))
            return &Pages[i];
        i++;
    }
    return NULL;
}

static tPSFONT *findfont(tSTYLE *style)
{
    int     i;

    i = 0;
    while (i < Fntcnt)
    {
        if (!strcmp(style->family, Fonts[i].id)
        &&  Fonts[i].attr == (style->attr & AF_FNTMSK)
        &&  (Fonts[i].points == -1 || Fonts[i].points == style->points))
            return &Fonts[i];
        i++;
    }
    return NULL;
}


static int chrwid(tFONT *f, unsigned char c)
{
    // @@@@ Kerning beruecksichtigen!
    if (f->prop)
        return f->widtab[c];
    else
        return f->x_dots;
}

static int dev_cfg_device(void)
{
    int     typ;
    char    typnam[L_NAME+1];

//    if (Dev.typ != DT_NONE)
//        return dev_error(devErrBAD, "more than one %Device");
    skipgap();
    if (!getname(Dev.id, L_ID, 0))
        return dev_error(devErrBAD, "device id expected");
    skipgap();
    if (!getname(typnam, L_NAME, 1))
        return dev_error(devErrBAD, "type expected");
//    typ = getdevtyp(typnam);
//    if (typ == DT_NONE)
//        return dev_error(devErrBAD, "invalid type");
//    Dev.typ = typ;
    return TRUE;
}

static int dev_cfg_afmpath(void)
{
    skipgap();
    if (*Cp == 0)
        return dev_error(devErrBAD, "afmpath expected");
    strcpy(Afmdir, Cp);
    return TRUE;
}

static int dev_cfg_page(void)
{
    char    opt[L_ID+1];

    if (Pagcnt >= MAXPAGES)
        return dev_error(devErrBAD, "too many pages");
    skipgap();
    if (!getname(Pages[Pagcnt].id, L_ID, 0))
        return dev_error(devErrBAD, "page-id expected");
    skipgap();
    if (!getunit(&Pages[Pagcnt].xlim))
        return dev_error(devErrBAD, "xlim expected");
    skipgap();
    if (!getunit(&Pages[Pagcnt].ylim))
        return dev_error(devErrBAD, "ylim expected");
    skipgap();
    if (!getunit(&Pages[Pagcnt].lm))
        return dev_error(devErrBAD, "lm expected");
    skipgap();
    if (!getunit(&Pages[Pagcnt].rm))
        return dev_error(devErrBAD, "rm expected");
    skipgap();
    if (!getunit(&Pages[Pagcnt].tm))
        return dev_error(devErrBAD, "tm expected");
    skipgap();
    if (!getunit(&Pages[Pagcnt].bm))
        return dev_error(devErrBAD, "bm expected");
    /* set default values for the optional parameters */
    Pages[Pagcnt].rotate = 0;
    Pages[Pagcnt].xshift = 0;
    Pages[Pagcnt].yshift = 0;
    Pages[Pagcnt].xscale = 1000;
    Pages[Pagcnt].yscale = 1000;
    /* parse the options */
    skipgap();
    while (needdot())
    {
        if (!getname(opt, L_ID, 0))
            return dev_error(devErrBAD, "option expected");
        if (!needchar('='))
            return dev_error(devErrBAD, "'=' expected");
        if (!strcmp(opt, "rotate")) {
            if (!getdec(&Pages[Pagcnt].rotate))
                return dev_error(devErrBAD, "rotation angle expected");
        } else if (!strcmp(opt, "shift")) {
            if (!getunit(&Pages[Pagcnt].xshift))
                return dev_error(devErrBAD, "x-offset expected");
            if (!needcomma())
                return dev_error(devErrBAD, "comma expected");
            if (!getunit(&Pages[Pagcnt].yshift))
                return dev_error(devErrBAD, "y-offset expected");
        } else if (!strcmp(opt, "scale")) {
            if (!getdec(&Pages[Pagcnt].xscale))
                return dev_error(devErrBAD, "x-scaling factor expected");
            if (!needcomma())
                return dev_error(devErrBAD, "comma expected");
            if (!getdec(&Pages[Pagcnt].yscale))
                return dev_error(devErrBAD, "y-scaling factor expected");
        } else {
            return dev_error(devErrBAD, "invalid option");
        }
        skipgap();
    }
    Pagcnt++;
    return TRUE;
}

static int dev_cfg_psfont(void)
{
    int      i;
    char     c, filename[8+1];
    tPSFONT  *font;

    if (Fntcnt >= MAXFONTS)
        return dev_error(devErrBAD, " too many fonts");
    font = &Fonts[Fntcnt];
    skipgap();
    if (!getname(font->id, L_ID, FALSE))
        return dev_error(devErrBAD, "font-id expected");
    skipgap();
    if (*Cp == '*')
    {
        font->points = -1;
        Cp++;
    }
    else
        if (!getdec(&(font->points)))
            return dev_error(devErrBAD, "points expected");
    skipgap();
    if (!isalpha(*Cp))
        return dev_error(devErrBAD, "attributes expected");
    font->attr = 0;
    while (isalpha(*Cp))
    {
        switch (*Cp)
        {
            case 'N':   font->attr  = 0;         break;
            case 'B':   font->attr |= AF_BOLD;   break;
            case 'I':   font->attr |= AF_ITALIC; break;
            default:    return dev_error(devErrBAD, "invalid attribute");
        }
        Cp++;
    }
    font->p_flag = FALSE;
    for (i=0; i<256; i++)
        font->widtab[i] = 0;
    font->upos = 0; font->uwid = 0;
    font->bbx0 = 0; font->bby0 = 0; font->bbx1 = 0; font->bby1 = 0;
    font->kerncnt = 0;
    font->kerntab = NULL;
    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!getname(filename, 8, FALSE))
            return dev_error(devErrBAD, "filename expected");
        if (!scan_afmfile(font, filename))
            return FALSE;
        font->p_flag = TRUE;
    }
    else
    {
        if (!getdec(&(font->relwidth)))
            return dev_error(devErrBAD, "relwidth expected");
        skipgap();
        if (*Cp == 0)
            return dev_error(devErrBAD, "PS-fontname expected");
        strcpy(font->fontname, Cp);
    }
    font->refnr = 0;
    Fntcnt++;
    return TRUE;
}

static int dev_cfg_adjust(void)
{
    skipgap();
    if (!getunit(&Xadj))
        return dev_error(devErrBAD, "x-adjust expected");
    skipgap();
    if (!getunit(&Yadj))
        return dev_error(devErrBAD, "y-adjust expected");
    return TRUE;
}

static int dev_cfg_item(void)
{
    char    buf[L_NAME+1];

    if (!getname(buf, L_NAME, 0))
        return dev_error(devErrBAD, "keyword expected");
    if (!strcmp(buf, "Device"))
        return dev_cfg_device();
    if (!strcmp(buf, "AfmPath"))
        return dev_cfg_afmpath();
    if (!strcmp(buf, "Adjust"))
        return dev_cfg_adjust();
    if (!strcmp(buf, "Page"))
        return dev_cfg_page();
    if (!strcmp(buf, "PsFont"))
        return dev_cfg_psfont();
    return dev_error(devErrBAD, "unknown keyword");
}

static  char    cfgline[256];
int DevConfig(char *filnam)
{
    FILE    *fp;
    char    fn[L_NAME+1], c;

    verbose(1, "device-configuration...\n");
    verbose(2, "  filnam='%s', Usrdir='%s', Lpudir='%s'\n",
        filnam, Usrdir, Lpudir);
    sprintf(fn, "%s", filnam);
    fp = fopen(fn, "r");
    if (fp == NULL && *fn != '/')
    {
        sprintf(fn, "%s/.lpu/%s", Usrdir, filnam);
        fp = fopen(fn, "r");
        if (fp == NULL)
        {
            sprintf(fn, "%s/%s", Lpudir, filnam);
            fp = fopen(fn, "r");
            if (fp == NULL)
                return dev_error(devErrPAR, "no device-file '%s'",filnam);
        }
    }
    verbose(1, "using devfile '%s' ...\n", fn);
    while (nextline(fp, cfgline, 255))
    {
        skipgap();
        switch (*Cp)
        {
            case 0:
            case ';':
                break;
            case '%':
                Cp++;
                dev_cfg_item();
                break;
            default:
                return dev_error(devErrBAD, "invalid line format");
        }
    }
    fclose(fp);
    return devOK;
}

static int gsaved = 0;

static void gsave(void)
{
    fprintf(Ofp, "gsave\n");
    gsaved++;
}

static void grestore(void)
{
    if (gsaved <= 0)
        return;
    fprintf(Ofp, "grestore\n");
    gsaved--;
}

static int curattr = 0;

static void set_attr(int attr)
{
    if (attr == curattr)
        return;
    fprintf(Ofp, "SA ");
    if (attr == 0)
    {
        fprintf(Ofp, "N\n");
        return;
    }
    if (attr & AF_BOLD)
        fprintf(Ofp, "B");
    if (attr & AF_ULINE)
        fprintf(Ofp, "U");
    if (attr & AF_ITALIC)
        fprintf(Ofp, "I");
    fprintf(Ofp, "\n");
    curattr = attr;
}

static void set_dashed(int dash)
{
    switch (dash)
    {
        case 0:     /* solid */
            fprintf(Ofp, " [] 0 setdash\n");
            break;
        case 1:     /* dashed */
            fprintf(Ofp, " [4 4] 0 setdash\n");
            break;
        case 2:     /* dotted */
            fprintf(Ofp, " [2 6] 0 setdash\n");
            break;
        case 3:     /* dashed and dotted */
            fprintf(Ofp, " [3 2 1 2] 0 setdash\n");
            break;
        default:
            break;
    }
}

static int calc_x(int x)
{
    return (Xadj + Xbase + x + 5) / 10;
}

static int calc_y(int y)
{
    return (Ymax - Ybase - Yadj - y + 5) / 10;
}


static void prt_pos(void)
{
    fprintf(Ofp, "%d %d M ", calc_x(Xpos), calc_y(Ypos));
}


static void labelframe(int frame)
{
    grestore();
    gsave();
    fprintf(Ofp, "%d %d moveto\n", calc_x(0),    calc_y(0)   );
    fprintf(Ofp, "%d %d lineto\n", calc_x(0),    calc_y(Ysiz));
    fprintf(Ofp, "%d %d lineto\n", calc_x(Xsiz), calc_y(Ysiz));
    fprintf(Ofp, "%d %d lineto\n", calc_x(Xsiz), calc_y(0)   );
    fprintf(Ofp, "closepath clip\n");
    if (OptFrame || frame)
        fprintf(Ofp, "0.1 setlinewidth stroke ");
    fprintf(Ofp, "newpath\n");
    CurrentPen = -1;
}


int DevReset(void)
{
    Xadj     = 0;
    Yadj     = 0;
    Pagcnt   = 0;
    Fntcnt   = 0;
    return devOK;
}

tPAGE *DevSelPage(char *id)
{
    tPAGE  *p;

    pPag = findpage(id);
    if (pPag == NULL)
        return NULL;
    Xmax   = pPag->xlim;
    Ymax   = pPag->ylim;
    Lm     = pPag->lm;
    Rm     = Xmax - pPag->rm;
    Tm     = pPag->tm;
    Bm     = Ymax - pPag->bm;
    Rotate = pPag->rotate;
    Xshift = pPag->xshift;
    Yshift = pPag->yshift;
    Xscale = pPag->xscale;
    Yscale = pPag->yscale;
    Xbase  = 0;
    Ybase  = 0;
    Xsiz   = Xmax;
    Ysiz   = Ymax;
    Xpos   = 0;
    Ypos   = 0;
    fprintf(Ofp, "%% Layout = %s\n",pPag->id);
    return pPag;
}


int DevGetFontCount(void)
{
    return Fntcnt;
}

int DevGetFontInfo(int fnr, tFONT *font)
{
    if ((fnr < 0) || (fnr >= Fntcnt))
        return dev_error(devErrBAD, "invalid font index");
    memcpy(font, &Fonts[fnr], sizeof(tFONT));
    return 1;
}

int DevSelFont(tSTYLE *style)
{
    int     i, points, diff;
    double  x;

    //if (curattr != 0)
    //    set_attr(0);
    if (style == NULL)
    {
        pFnt = &Fonts[0];
        points = 12;
    }
    else
    {
        pFnt = findfont(style);
        if (pFnt == NULL)
            return 0;
        points = style->points;
    }
    // @@@ set_attr(pFnt->defattr);
    strcpy(Myfont.id, pFnt->id);
    Myfont.points = points;
    Myfont.prop   = pFnt->p_flag;
    Myfont.attr   = pFnt->attr;
    Myfont.y_dots = Myfont.points * 10;
    Myfont.x_dots = Myfont.points * pFnt->relwidth / 100;
    diff          = pFnt->bby1 - pFnt->bby0;
    x             = 0;
    if (diff > 0)
        x         = pFnt->bby1 / diff;
    if (x <= 0.1 || x >= 1.)
        x         = 0.75;
    Myfont.b_dots = Myfont.y_dots * x;
    Myfont.upos   = pFnt->upos * Myfont.points / 100;
    Myfont.uwid   = pFnt->uwid * Myfont.points / 100;
    if (Myfont.prop)
    {
        for (i=0; i<256; i++)
            Myfont.widtab[i] = Myfont.points * pFnt->widtab[i] / 100;
    }
    if (pFnt->refnr == 0)
    {
        pFnt->refnr = ++Fntref;
        fprintf(Ofp, "/F%d /%s ISO\n", pFnt->refnr, pFnt->fontname);
    }
    fprintf(Ofp, "%d /F%d FONT\n", points, pFnt->refnr);
    return 1;
}

int DevGetFontXdots(void)
{
    return Myfont.x_dots;
}

int DevGetFontYdots(void)
{
    return Myfont.y_dots;
}

int DevGetFontBdots(void)
{
    return Myfont.b_dots;
}

int DevGetFontIsProp(void)
{
    return Myfont.prop;
}

int DevSetAttr(int attr)
{
    set_attr(attr);
    return devOK;
}

int DevBeginJob(FILE *fp)
{
    Ofp = fp;
    fprintf(Ofp, "%%!PS-Adobe-2.0\n");
    fprintf(Ofp, "%%%%Creator: LPU V. %s\n", Version);
    fprintf(Ofp, "%%%%Title: output from devapi\n");
    //fprintf(Ofp, "%%%%BoundingBox %d %d %d %d\n", Lm, Bm, Xmax-Rm, Ymax-Tm);
    fprintf(Ofp, "%%%%BoundingBox 0 0 595 842\n");
    fprintf(Ofp, "%%%%EndComment\n");
    fprintf(Ofp, "%%%%EndProlog\n");
    fprintf(Ofp, "\n");
    fprintf(Ofp, "/ISO { findfont dup length dict begin\n");
    fprintf(Ofp, "  { 1 index /FID ne { def } { pop pop } ifelse } forall\n");
    fprintf(Ofp, "  /Encoding ISOLatin1Encoding def currentdict end\n");
    fprintf(Ofp, "  definefont pop\n");
    fprintf(Ofp, "} bind def\n");
    fprintf(Ofp, "/FONT { findfont exch scalefont setfont } bind def\n");
    fprintf(Ofp, "/M { moveto } bind def\n");
    fprintf(Ofp, "/S { show } bind def\n");
    fprintf(Ofp, "/SU { currentpoint 2 index stringwidth dup 0 exch 2 sub "
                      "rmoveto rlineto stroke moveto show } bind def\n");
    Pagnr = 0;
    return devOK;
}

int DevEndJob(void)
{
    if (curattr != 0)
        set_attr(0);
    fprintf(Ofp, "%%%%Trailer\n");
    fprintf(Ofp, "%%%%EOF\n");
    return devOK;
}

static int newpage = 1;

int DevBeginPage(void)
{
    double  xs, ys;

    Xpos = 0;
    Ypos = 0;
    if (newpage)
    {
        Pagnr++;
        fprintf(Ofp, "%%%%Page %d %d\n", Pagnr, Pagnr);
    }
    gsave();  // 20010102
    /* scale, rotate and shift the page, if required */
    if (Xscale != 1000 || Yscale != 1000)
    {
        xs = Xscale / 1000.;
        ys = Yscale / 1000.;
        fprintf(Ofp, "%lf %lf scale\n", xs, ys);
    }
    if (Xshift != 0 || Yshift != 0)
        fprintf(Ofp, "%d %d translate\n", Xshift/10, Yshift/10);
    if (Rotate > 0)
        fprintf(Ofp, "%d rotate\n", Rotate);
    gsave();
    return devOK;
}

int DevEndPage(int show)
{
    grestore();
    grestore(); // 20010102
    if (show)
        fprintf(Ofp, "showpage\n");
    newpage = show;
    return devOK;
}

int DevSetWindow(int xpos, int ypos, int xsiz, int ysiz, int frame)
{
    if (xpos < 0 || xpos >= Rm)
        return dev_error(devErrPAR, "invalid x-position");
    if (ypos < 0 || ypos >= Bm)
        return dev_error(devErrPAR, "invalid y-position");
    if (xsiz <= 0 || xpos + xsiz > Rm)
        return dev_error(devErrPAR, "invalid x-size");
    if (ysiz <= 0 || ypos + ysiz > Bm)
        return dev_error(devErrPAR, "invalid y-size");
    Xbase = xpos;
    Ybase = ypos;
    Xsiz  = xsiz;
    Ysiz  = ysiz;
    Xpos  = 0;
    Ypos  = 0;
    verbose(1, "DevSetWin: Xbase=%d, Xsiz=%d, Ybase=%d, Ysiz=%d\n",
        Xbase, Xsiz, Ybase, Ysiz);
    labelframe(frame);
    return devOK;
}

int DevClearWindow(void)
{
    Xbase = 0;
    Ybase = 0;
    Xsiz  = Xmax;
    Ysiz  = Ymax;
    Xpos  = 0;
    Ypos  = 0;
    return devOK;
}

int DevExit(void)
{
    return devOK;
}

int DevSetXY(int x, int y)
{
    if (x < 0 || x >= Xsiz)
        return dev_error(devErrPAR, "invalid x-position");
    if (y < 0 || y >= Ysiz)
        return dev_error(devErrPAR, "invalid y-position");
    Xpos = x;
    Ypos = y;
    return devOK;
}

int DevAddY(int n)
{
    if (Ypos + n >= Ysiz)
        return dev_error(devErrPAR, "invalid offset");
    Ypos += n;
    return devOK;
}

int DevSetX(int x)
{
    if (x < 0 || x >= Xsiz)
        return dev_error(devErrPAR, "invalid position");
    Xpos = x;
    return devOK;
}

int DevGetChrWid(char c)
{
    return chrwid(&Myfont, c);
}

int DevGetStrWid(char *str)
{
    int     wid;
    
    wid = 0;
    while (*str)
        wid += chrwid(&Myfont, *str++);
    return wid;
}

//----- Zeichenfunktionen -----

static void adj_pen(void)
{
    float   wid;

    if (SelectedPen == CurrentPen)
        return;
    wid = (float) SelectedPen / 10;
    fprintf(Ofp, "%f setlinewidth\n", wid);
    CurrentPen = SelectedPen;
}

static void adj_ink(void)
{
    if (SelectedInk.cyan    == CurrentInk.cyan
    &&  SelectedInk.magenta == CurrentInk.magenta
    &&  SelectedInk.yellow  == CurrentInk.yellow
    &&  SelectedInk.black   == CurrentInk.black)
        return;
    fprintf(Ofp, "%.2f %.2f %.2f %.2f setcmykcolor\n",
        (float) SelectedInk.cyan    / 100.,
        (float) SelectedInk.magenta / 100.,
        (float) SelectedInk.yellow  / 100.,
        (float) SelectedInk.black   / 100.);
    memcpy(&CurrentInk, &SelectedInk, sizeof(tCMYK));
}

int DevRawStr(char *str)
{
    fprintf(Ofp, "%s", str);
    return devOK;
}

int DevPrtStr(char *str, int flags)
{
    unsigned char *p;

    p = str;
    adj_pen();
    adj_ink();
    if (strlen(p) > 0)
    {
        prt_pos();
        fprintf(Ofp, "(");
        while (*p)
        {
            if (*p < 32 || *p > 127 || *p == '(' || *p == ')')
                fprintf(Ofp, "\\%03o", *p);
            else
            {
                fprintf(Ofp, "%c", *p);
                if (*p == '\\')
                    fprintf(Ofp, "%c", *p);
            }
            p++;
        }
        if ((Myfont.attr & AF_ULINE) || (flags & AF_ULINE))
            fprintf(Ofp, ") SU\n");
        else
            fprintf(Ofp, ") S\n");
    }
    return devOK;
}

int DevDrawBox(int xsize, int ysize, int filled)
{
    adj_pen();
    adj_ink();
    /* auf die aktuelle Position gehen */
    fprintf(Ofp, "%d %d moveto\n", calc_x(Xpos),       calc_y(Ypos));
    /* Rechteck mit Breite 'xsize' und Hoehe 'ysize' aufziehen */
    fprintf(Ofp, "%d %d lineto\n", calc_x(Xpos),       calc_y(Ypos+ysize));
    fprintf(Ofp, "%d %d lineto\n", calc_x(Xpos+xsize), calc_y(Ypos+ysize));
    fprintf(Ofp, "%d %d lineto\n", calc_x(Xpos+xsize), calc_y(Ypos));
    /* und ausgeben. */
    fprintf(Ofp, "closepath");
    if (filled)
        fprintf(Ofp, " fill");
    fprintf(Ofp, " stroke\n");
    return devOK;
}

int DevDrawRline(int x, int y)
{
    adj_pen();
    adj_ink();
    /* auf die aktuelle Position gehen */
    fprintf(Ofp, "%d %d moveto\n", calc_x(Xpos), calc_y(Ypos));
    /* eine Linie zum relativen Punkt (x,y) ziehen */
    fprintf(Ofp, "%d %d lineto\n", calc_x(Xpos+x), calc_y(Ypos+y));
    /* und ausgeben. */
    fprintf(Ofp, "stroke\n");
    return devOK;
}

int DevDrawCircle(int r, int filled)
{
    adj_pen();
    adj_ink();
    /* auf die aktuelle Position gehen */
    fprintf(Ofp, "newpath %d %d %d 0 360 arc closepath",
        calc_x(Xpos), calc_y(Ypos), r/10);
    if (filled)
        fprintf(Ofp, " fill");
    fprintf(Ofp, " stroke\n");
    return devOK;
}

int DevSelPen(int pen)
{
    SelectedPen = pen;
    return devOK;
}

int DevSelCMYK(tCMYK *color)
{
    if (color == NULL)
        memcpy(&SelectedInk, &DefaultInk, sizeof(tCMYK));
    else
        memcpy(&SelectedInk, color, sizeof(tCMYK));
    return devOK;
}

int DevSelDash(int dash)
{
    if (dash < 0 || dash > 3)
        return dev_error(devErrPAR, "invalid dash index");
    Dashndx = dash;
    set_dashed(dash);
    return devOK;
}

/********** Support for EPS files **********/

static  FILE    *efp = NULL;
static  tPOS    epos = { 0, 0 };
static  tSIZE   esiz = { 1, 1 };
static  tSIZE   osiz = { 1, 1 };
static  int     bx0, by0, bx1, by1;
static  char    ebuf[1024];
static  fpos_t	fpos;

static int eps_boundingbox(void)
{
    skipgap();
    if (!getdec(&bx0))
        return dev_error(devErrBAD, "bounding box x0 expected");
    skipgap();
    if (!getdec(&by0))
        return dev_error(devErrBAD, "bounding box y0 expected");
    skipgap();
    if (!getdec(&bx1))
        return dev_error(devErrBAD, "bounding box x1 expected");
    skipgap();
    if (!getdec(&by1))
        return dev_error(devErrBAD, "bounding box y1 expected");
    return 1;
}

static int eps_translate()
{
    int     xnew = calc_x(epos.xpos);
    int     ynew = calc_y(epos.ypos+esiz.ysiz);
    int     tx = xnew - bx0, ty = ynew - by0;

    fprintf(Ofp, "%d %d translate\n", tx, ty);
    return 1;
}

static int eps_scale()
{
    double  xold = (bx1 - bx0) * 10, yold = (by1 - by0) * 10;
    double  xnew = esiz.xsiz, ynew = esiz.ysiz;
    double  xs = xnew / xold, ys = ynew / yold;

    fprintf(Ofp, "%f %f scale\n", xs, ys);
    return 1;
}

static int eps_prefix(void)
{
    fprintf(Ofp, "/LPUSave save def\n");
    fprintf(Ofp, "userdict 200 dict begin\n");
    fprintf(Ofp, "/showpage {} def /copypage {} def /erasepage {} def\n");
    fprintf(Ofp, "gsave\n");
    eps_translate();
    eps_scale();
    return 1;
}

static int eps_suffix(void)
{
    fprintf(Ofp, "grestore\n");
    fprintf(Ofp, "end\n");
    fprintf(Ofp, "LPUSave restore\n");
    fprintf(Ofp, "\n");
    return 1;
}

int DevEPSOpen(char *fn, tSIZE *size)
{
    int     mode = 0;

    if (efp != NULL)
        return dev_error(devErrBAD, "EPS file still open");
    efp = fopen(fn, "r");
    if (efp == NULL)
        return dev_error(devErrBAD, "unknown filename '%s'", fn);
    while (!feof(efp) && mode < 3)
    {
        fgets(ebuf, 1023, efp);
        if (feof(efp))
            continue;
        Cp = ebuf;
        switch (mode)
        {
        case 0:
            if (memcmp(Cp, "%!", 2))
                return error("%! expected");
            mode = 1;
            break;
        case 1:
            if (memcmp(Cp, "%", 1))
                return error("% expected");
            if (memcmp(Cp, "%%BoundingBox:", 14))
                break;
            Cp += 14;
            if (!eps_boundingbox())
                return FALSE;
            mode = 2;
            break;
        case 2:
            if (!memcmp(Cp, "%", 1))
                break;
            fgetpos(efp, &fpos);
            mode = 3;
            break;
        }
    }
    if (mode != 3)
        return dev_error(devErrBAD, "file contains no data");
    osiz.xsiz = (bx1 - bx0) * 10;
    osiz.ysiz = (by1 - by0) * 10;
    esiz.xsiz = osiz.xsiz;
    esiz.ysiz = osiz.ysiz;
    epos.xpos = 0;
    epos.ypos = 0;
    size->xsiz = esiz.xsiz;
    size->ysiz = esiz.ysiz;
    return 1;
}

int DevEPSResize(tSIZE *size)
{
    esiz.xsiz = size->xsiz;
    esiz.ysiz = size->ysiz;
    // @@@ check!
    return 1;
}

int DevEPSRescale(tSCALE *scale)
{
    esiz.xsiz = (int) (osiz.xsiz * scale->xscale);
    esiz.ysiz = (int) (osiz.ysiz * scale->yscale);
}

int DevEPSAt(tPOS *pos)
{
    epos.xpos = pos->xpos;
    epos.ypos = pos->ypos;
    // @@@ check!
    return 1;
}

int DevEPSPrint(void)
{
    char    buf[256];

    if (efp == NULL)
        dev_error(devErrBAD, "eps file not open");
    eps_prefix();
    rewind(efp);
    fgetpos(efp, &fpos);
//    fsetpos(efp, &fpos);
//    fprintf(Ofp, "%s", ebuf); /* first line of body already in buffer */
    while (!feof(efp))
    {
        fgets(ebuf, 1023, efp);
        if (feof(efp))
            continue;
        fprintf(Ofp, "%s", ebuf);
    }
    eps_suffix();
    return 1;
}

int DevEPSClose(void)
{
    if (efp == NULL)
        dev_error(devErrBAD, "eps file not open");
    fclose(efp);
    efp = NULL;
    return 1;
}

int DevEPSInUse(void)
{
    if (efp == NULL)
        return 0;
    return 1;
}

int DevEPSGetSize(tSIZE *size)
{
    size->xsiz = esiz.xsiz;
    size->ysiz = esiz.ysiz;
    return 1;
}

