/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 *      devapi.h --    Definitionen fuer die DEV-Treiber
 *
 *          jrs     13.07.97
 */

#ifndef devapi_h
#define devapi_h


#define AF_ULINE    0x0001
#define AF_BOLD     0x0002
#define AF_ITALIC   0x0004

#define AF_FNTMSK   (AF_BOLD|AF_ITALIC)
#define AF_MASK     (AF_ULINE|AF_BOLD|AF_ITALIC)

#define MAXPAGES    25
#define MAXFONTS    50

#define devOK       0
#define devErrBAD   -1
#define devErrFCT   -2
#define devErrPAR   -3
#define devErrNON   -4
#define devErrILL   -5

typedef struct      /* Geraetebezogene Daten                            */
{
    char    id[L_ID+1];         /* Device-ID                */
    int     typ;                /* Device-Type              */
    char    name[L_NAME+1];     /* Device-Name              */
    int     xadj;               /* Korrektur X              */
    int     yadj;               /* Korrektur Y              */
} tDEV;

typedef struct      /* Seitenbeschreibung                               */
{
    char    id[L_ID+1];
    int     xlim;
    int     ylim;
    int     lm;
    int     rm;
    int     tm;
    int     bm;
    int     rotate;
    int     xshift;
    int     yshift;
    int     xscale;
    int     yscale;
} tPAGE;

typedef struct      /* Parameter des aktuellen Zeichensatzes            */
{
    char    id[L_ID+1];         /* aktueller Fontname                   */
    int     points;             /* aktuelle Punktgroesse                */
    int     prop;               /* Flag: proportional oder fixed?       */
    int     attr;
    int     y_dots;             /* aktuelle Hoehe in DOTs               */
    int     x_dots;             /* aktuelle Standardbreite in DOTs      */
    int     b_dots;             /* baseline position in DOTs (from top) */
    int     upos;               /* underline position in DOTs           */
    int     uwid;               /* underline width in DOTs              */
    int     widtab[256];        /* aktuelle Breitentab. in DOTs         */
} tFONT;

typedef struct
{
    unsigned char   first;
    unsigned char   second;
    short           shift;
} tKPX;

typedef struct      /* Informationen ueber den Zeichensatz              */
{
    char    id[L_ID+1];         /* Font-ID                      */
    char    fontname[L_NAME+1]; /* internal fontname            */
    int     points;             /* Punktgroesse (-1 fuer 'alle')*/
    int     refnr;              /* Referenznummer, wenn > 0     */
    int     p_flag;             /* flag: proportional font      */
    int     relwidth;           /* char-width in 1/1000 height  */
    int     attr;               /* font attributes              */
    int     widtab[256];        /* width-table [0..255]         */
    int     upos;               /* underline position           */
    int     uwid;               /* underline thickness          */
    int     bbx0, bby0;         /* bounding box x0/y0           */
    int     bbx1, bby1;         /* bounding box x1/y1           */
    int     kerncnt;            /* nbr of entries in the kerntab*/
    tKPX   *kerntab;            /* table of kerning pairs       */
} tPSFONT;

int     dev_error(int err, char *fmt, ...);

int     DevReset(void);
int     DevConfig(char *filnam);
tPAGE  *DevSelPage(char *id);

int     DevGetFontCount(void);
int     DevGetFontInfo(int fnr, tFONT *font);
int     DevSelFont(tSTYLE *style);
int     DevGetFontXdots(void);
int     DevGetFontYdots(void);
int     DevGetFontBdots(void);
int     DevGetFontIsProp(void);

int     DevBeginJob(FILE *ofp);
int     DevEndJob(void);
int     DevBeginPage(void);
int     DevEndPage(int show);
int     DevSetWindow(int xpos, int ypos, int xsiz, int ysiz, int frame);
int     DevClearWindow(void);
int     DevExit(void);
int     DevSetXY(int x, int y);
int     DevSetX(int x);
int     DevAddY(int n);
int     DevSetAttr(int attr);
int     DevGetChrWid(char c);
int     DevGetStrWid(char *str);
int     DevPrtStr(char *str, int flags);
int     DevDrawBox(int xsize, int ysize, int filled);
int     DevDrawRline(int x, int y);
int     DevDrawCircle(int r, int filled);
int     DevSelPen(int pen);
int     DevSelCMYK(tCMYK *color);
int     DevSelDash(int dash);

int     DevEPSOpen(char *fn, tSIZE *size);
int     DevEPSResize(tSIZE *size);
int     DevEPSRescale(tSCALE *scale);
int     DevEPSAt(tPOS *pos);
int     DevEPSPrint(void);
int     DevEPSClose(void);
int     DevEPSInUse(void);
int     DevEPSGetSize(tSIZE *size);

#endif	/* devapi_h */
