/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 *  expr.c -- Berechnung von Ausdruecken
 *
 *      jrs     02.08.97
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "cfg.h"
#include "utils.h"
#include "expr.h"
#include "lpudefs.h"
#include "lpuvars.h"
#include "devapi.h"
#include "tdbfuncs.h"

extern  int     Frmcnt;
extern  int     Frmndx;
extern  tFORM   *Form;
extern  tFORM   Forms[];
extern  tLABEL  Labels[];

VALUE   Rc;

//--------------------------------------------

SYMBOL *symFind(tVARS *blk, char *id)
{
    SYMBOL  *sym;

    sym = blk->first;
    while (sym != NULL)
    {
        if (!strcmp(id, sym->name))
            return sym;
        sym = sym->next;
    }
    return NULL;
}

SYMBOL  *symFindLoc(char *id)
{
    SYMBOL  *sym;

    if (Ctxcnt >= 0)
    {
        sym = symFind(&Ctxtab[Ctxcnt].vars, id);
        return sym;
    }
    return NULL;
}

SYMBOL  *symFindAll(char *id)
{
    int     n;
    SYMBOL  *sym;

    n = Ctxcnt;
    while (n >= 0)
    {
        sym = symFind(&Ctxtab[n].vars, id);
        if (sym != NULL)
            return sym;
        n--;
    }
    sym = symFind(&Gsyms, id);
    if (sym != NULL)
        return sym;
    return NULL;
}

SYMBOL	*symAdd(tVARS *blk, char *id)
{
    SYMBOL  *sym;

    sym = (SYMBOL *) calloc(1, sizeof(SYMBOL));
    if (sym == NULL)
        return NULL;
    sym->next    = NULL;
    strcpy(sym->name, id);
    sym->val.typ = T_NUM;
    sym->val.str[0] = 0;
    if (blk->first == NULL)
    {
        blk->first = sym;
        blk->last  = sym;
    }
    else
    {
        blk->last->next = sym;
        blk->last       = sym;
    }
    return sym;
}

void symDelAll(tVARS *blk)
{
    SYMBOL  *sym, *p;

    sym = blk->first;
    while (sym != NULL)
    {
        p = sym;
        sym = sym->next;
        free(p);
    }
    blk->first = NULL;
    blk->last  = NULL;
}

static char *symTypToName(int typ)
{
    if (typ == T_NUM)
        return "T_NUM";
    if (typ == T_STR)
        return "T_STR";
    if (typ == T_FLT)
        return "T_FLT";
    return "???";
}

void symDumpAll(tVARS *blk)
{
    SYMBOL  *sym;

    fprintf(stderr, "Dump Symbols\n");
    sym = blk->first;
    while (sym != NULL)
    {
        fprintf(stderr, "  Symbol '%s': typ=%d,'%s', num=%8ld, str='%s'\n",
            sym->name, sym->val.typ, symTypToName(sym->val.typ),
            sym->val.num, sym->val.str);
        sym = sym->next;
    }
}

static int to_num(VALUE *v)
{
    if (v->typ == T_NUM)
        return 1;
    if (v->typ == T_STR)
    {
        if (!isdigit(v->str[0]))
            return 0;
        v->num = atoi(v->str);
        v->typ = T_NUM;
        return 1;
    }
    if (v->typ == T_FLT)
    {
        v->num = v->flt;
        v->typ = T_NUM;
        return 1;
    }
    return 0;
}

int number(VALUE *v)
{
    double  dec;

    //@@@ vorerst nur Dezimal
    v->typ = T_NUM;
    v->num = 0;
    dec  = 1;
    while (isdigit(*Cp))
        v->num = v->num * 10 + (*Cp++ - '0');
    if (*Cp == '.')
    {
        v->typ = T_FLT;
        v->flt = (double) v->num;
        Cp++;
        while (isdigit(*Cp))
        {
            dec /= 10.;
            v->flt = v->flt + dec * (*Cp++ - '0');
        }
    }
    return 1;
}


int symbol(VALUE *v)
{
    char    buf[L_NAME+1];
    SYMBOL  *sym;

    if (!getname(buf, L_NAME, 0))
        return 0;
    sym = symFindAll(buf);
    if (sym == NULL)
        return 0;
    v->typ = sym->val.typ;
    v->num = sym->val.num;
    v->flt = sym->val.flt;
    strcpy(v->str, sym->val.str);
    return 1;
}

/***** SystemVariablen *****/

int sv_now(VALUE *v)
{
    time_t  t;

    t = time(NULL);
    v->num = (int) t;
    v->typ = T_NUM;
    return 1;
}

int sv_count(VALUE *v)
{
    v->typ = T_NUM;
    v->num = Cnt.num;
    return 1;
}

int sv_parcnt(VALUE *v)
{
    v->typ = T_NUM;
    v->num = pCtx->proc->parcnt;
    return 1;
}

int sv_formcnt(VALUE *v)
{
    v->typ = T_NUM;
    v->num = Frmcnt;
    return 1;
}

int sv_fontcnt(VALUE *v)
{
    v->typ = T_NUM;
    v->num = DevGetFontCount();
    return 1;
}

int sv_propfont(VALUE *v)
{
    v->typ = T_NUM;
    v->num = DevGetFontIsProp();
    return 1;
}

int sv_fontheight(VALUE *v)
{
    v->typ = T_NUM;
    v->num = DevGetFontYdots();
    return 1;
}

int sv_baseline(VALUE *v)
{
    v->typ = T_NUM;
    v->num = DevGetFontBdots();
    return 1;
}

int sv_xbase(VALUE *v)
{
    v->typ = T_NUM;
    v->num = W->xoff;
    return 1;
}

int sv_ybase(VALUE *v)
{
    v->typ = T_NUM;
    v->num = W->yoff;
    return 1;
}

int sv_xsize(VALUE *v)
{
    v->typ = T_NUM;
    v->num = W->xsiz;
    return 1;
}

int sv_ysize(VALUE *v)
{
    v->typ = T_NUM;
    v->num = W->ysiz;
    return 1;
}

int sv_xpos(VALUE *v)
{
    v->typ = T_NUM;
    v->num = W->xpos;
    return 1;
}

int sv_ypos(VALUE *v)
{
    v->typ = T_NUM;
    v->num = W->ypos;
    return 1;
}

int sv_version(VALUE *v)
{
    v->typ = T_STR;
    strcpy(v->str, Version);
    return 1;
}

int sv_gendate(VALUE *v)
{
    v->typ = T_STR;
    strcpy(v->str, GENDATE);
    return 1;
}

int sv_batchmode(VALUE *v)
{
    v->typ = T_NUM;
    v->num = OptBatch;
    return 1;
}

/***** EPS information *****/

static int sv_eps_xsize(VALUE *v)
{
    tSIZE   size;

    v->typ = T_NUM;
    if (!DevEPSGetSize(&size))
        v->num = 1;
    else
        v->num = size.xsiz;
    return 1;
}

static int sv_eps_ysize(VALUE *v)
{
    tSIZE   size;

    v->typ = T_NUM;
    if (!DevEPSGetSize(&size))
        v->num = 1;
    else
        v->num = size.ysiz;
    return 1;
}

int sv_eps(VALUE *v)
{
    char    id[11];

    if (!needchar('{'))                 return error("'{' expected");
    if (!getname(id, 10, FALSE))        return error("ID expected");
    if (!needchar('}'))                 return error("'}' expected");

    if (!strcmp(id, "xsize"))           return sv_eps_xsize(v);
    if (!strcmp(id, "ysize"))           return sv_eps_ysize(v);
    return error("unknown id '%s'", id);
}

/***** System-Funktionen *****/

int sf_param(VALUE *v)
{
    VALUE    *p;

    if (!needlpar())            return 0;
    if (!num_expr(v))           return 0;
    if (!needrpar())            return 0;
    if (v->num < 0 || v->num > pCtx->proc->parcnt)
        return error("invalid parcnt");
    if (v->num == 0)
    {
        v->typ = T_STR;
        strcpy(v->str, pCtx->proc->procname);
        return 1;
    }
    p = pCtx->proc->params;
    p = p + v->num - 1;
    v->typ = p->typ;
    switch (v->typ) {
        case T_NUM:     v->num = p->num;                        break;
        case T_STR:     strcpy(v->str, p->str);                 break;
        default:                                                break;
    }
    return 1;
}

int sf_formid(VALUE *v)
{
    if (!needlpar())            return 0;
    if (!num_expr(v))           return 0;
    if (!needrpar())            return 0;
    if (v->num < 0 || v->num >= Frmcnt)
        return error("invalid formcnt");
    strcpy(v->str, Forms[v->num].id);
    v->typ = T_STR;
    return 1;
}

int sf_lblcnt(VALUE *v)
{
    int     ndx = 0;

    if (!needlpar())            return 0;
    if (needchar('.'))
        ndx = Frmndx;
    else
    {
        if (!num_expr(v))       return 0;
        ndx = v->num;
    }
    if (!needrpar())            return 0;
    if (ndx < 0 || ndx >= Frmcnt)
        return error("invalid formcnt");
    v->num = Forms[ndx].lblcnt;
    v->typ = T_NUM;
    return 1;
}

int sf_lblid(VALUE *v)
{
    int     fn = 0, ln;

    if (!needlpar())            return 0;
    if (needchar('.'))
        fn = Frmndx;
    else
    {
        if (!num_expr(v))       return 0;
        fn = v->num;
    }
    if (!needcomma())           return 0;
    if (!num_expr(v))           return 0;
    ln = v->num;
    if (!needrpar())            return 0;
    if (fn < 0 || fn >= Frmcnt)
        return error("invalid formcnt");
    if (ln < 0 || ln >= Forms[fn].lblcnt)
        return error("invalid labelcnt");
    strcpy(v->str, Labels[Forms[fn].lbl1st+ln].id);
    v->typ = T_STR;
    return 1;
}

int sf_lblxcnt(VALUE *v)
{
    int     fn = 0, ln;

    if (!needlpar())            return 0;
    if (needchar('.'))
        fn = Frmndx;
    else
    {
        if (!num_expr(v))       return 0;
        fn = v->num;
    }
    if (!needcomma())           return 0;
    if (!num_expr(v))           return 0;
    ln = v->num;
    if (!needrpar())            return 0;
    if (fn < 0 || fn >= Frmcnt)
        return error("invalid formcnt");
    if (ln < 0 || ln >= Forms[fn].lblcnt)
        return error("invalid labelcnt");
    v->num = Labels[Forms[fn].lbl1st+ln].xcnt;
    v->typ = T_NUM;
    return 1;
}

int sf_lblycnt(VALUE *v)
{
    int     fn = 0, ln;

    if (!needlpar())            return 0;
    if (needchar('.'))
        fn = Frmndx;
    else
    {
        if (!num_expr(v))       return 0;
        fn = v->num;
    }
    if (!needcomma())           return 0;
    if (!num_expr(v))           return 0;
    ln = v->num;
    if (!needrpar())            return 0;
    if (fn < 0 || fn >= Frmcnt)
        return error("invalid formcnt");
    if (ln < 0 || ln >= Forms[fn].lblcnt)
        return error("invalid labelcnt");
    v->num = Labels[Forms[fn].lbl1st+ln].ycnt;
    v->typ = T_NUM;
    return 1;
}

int sf_max(VALUE *v)
{
    VALUE   e;

    if (!needlpar())            return 0;
    if (!expr(v))               return 0;
    if (!needcomma())           return 0;
    if (!expr(&e))              return 0;
    if (!needrpar())            return 0;
    if ((v->typ == T_NUM) && (e.typ == T_NUM))
    {
        if (e.num > v->num)
            v->num = e.num;
        return 1;
    }
    if ((v->typ == T_FLT) && (e.typ == T_FLT))
    {
        if (e.flt > v->flt)
            v->flt = e.flt;
        return 1;
    }
    return error("sf_max: invalid operand type");
}

int sf_min(VALUE *v)
{
    VALUE   e;

    if (!needlpar())            return 0;
    if (!expr(v))               return 0;
    if (!needcomma())           return 0;
    if (!expr(&e))              return 0;
    if (!needrpar())            return 0;
    if ((v->typ == T_NUM) && (e.typ == T_NUM))
    {
        if (e.num < v->num)
            v->num = e.num;
        return 1;
    }
    if ((v->typ == T_FLT) && (e.typ == T_FLT))
    {
        if (e.flt < v->flt)
            v->flt = e.flt;
        return 1;
    }
    return error("sf_min: invalid operand type");
}

int sf_strlen(VALUE *v)
{
    if (!needlpar())
        return 0;
    if (!str_expr(v))
        return 0;
    if (!needrpar())
        return 0;
    v->num = strlen(v->str);
    v->typ = T_NUM;
    return 1;
}

int sf_strpad(VALUE *v)
{
    VALUE   s, len;
    int     n;
    char    *p;

    if (!needlpar())            return 0;
    if (!str_expr(v))           return 0;
    if (!needcomma())           return 0;
    if (!str_expr(&s))          return 0;
    if (!needcomma())           return 0;
    if (!num_expr(&len))        return 0;
    if (!needrpar())            return 0;
    p = s.str;
    if (strlen(v->str) == 0)
        strcpy(v->str, " ");
    n = strlen(v->str);
    if (len.num > L_STRING)
        len.num = L_STRING;
    while (n < len.num)
    {
        if (*p == 0)
            p = s.str;
        v->str[n++] = *p++;
    } 
    if (strlen(v->str) > len.num)
        v->str[len.num] = 0;
    v->str[n] = 0;
    return 1;
}

int sf_concat(VALUE *v)
{
    VALUE   s;
    int     n;
    char    *p;

    if (!needlpar())
        return 0;
    if (!str_expr(v))
        return 0;
    while (needcomma())
    {
        if (!str_expr(&s))
            return 0;
        n = strlen(v->str);
        p = s.str;
        while (n < L_STRING && *p)
            v->str[n++] = *p++;
        v->str[n] = 0;
    }
    if (!needrpar())
        return 0;
    return 1;
}

int sf_num(VALUE *v)
{
    if (!needlpar())            return 0;
    if (!expr(v))               return 0;
    if (!needrpar())            return 0;
    switch (v->typ)
    {
    case T_NUM:
        break;
    case T_FLT:
        v->typ = T_NUM;
        v->num = v->flt;
        break;
    case T_STR:
        v->typ = T_NUM;
        v->num = atoi(v->str);
        break;
    default:
        return error("sf_num: invalid type");
    }
    return 1;
}

int sf_strstr(VALUE *v)
{
    VALUE   s;
    char    *p;

    if (!needlpar())            return 0;
    if (!str_expr(v))           return 0;
    if (!needcomma())           return 0;
    if (!str_expr(&s))          return 0;
    if (!needrpar())            return 0;
    p = strstr(v->str, s.str);
    if (p == NULL)
        v->num = -1;
    else
        v->num = p - v->str;
    v->typ = T_NUM;
    return 1;
}

int sf_strtok(VALUE *v)
{
    VALUE   s1, s2;
    char    *p;

    if (!needlpar())            return 0;
    if (!str_expr(&s1))         return 0;
    if (needcomma())
    {
        if (!str_expr(&s2))     return 0;
        if (!needrpar())        return 0;
        p = strtok(s1.str, s2.str);
    }
    else
    {
        if (!needrpar())        return 0;
        p = strtok(NULL, s1.str);
    }
    if (p == NULL)
        v->str[0] = 0;
    else
        strcpy(v->str, p);
    v->typ = T_STR;
    return 1;
}

int sf_strftime(VALUE *v)
{
    VALUE       s, t;
    struct tm   *tm;

    if (!needlpar())            return 0;
    if (!num_expr(&t))          return 0;
    if (!needcomma())           return 0;
    if (!str_expr(&s))          return 0;
    if (!needrpar())            return 0;
    tm = localtime(&(t.num));
    strftime(v->str, L_STRING, s.str, tm);
    v->typ = T_STR;
    return 1;
}

int sf_strseg(VALUE *v)
{
    VALUE   s, p, l;
    int     off;
    char    *str;

    if (!needlpar())            return 0;
    if (!str_expr(&s))          return 0;
    if (!needcomma())           return 0;
    if (!num_expr(&p))          return 0;
    if (!needcomma())           return 0;
    if (!num_expr(&l))          return 0;
    if (!needrpar())            return 0;
    if (p.num < 0)              return 0;
    v->str[0] = 0;
    v->typ = T_STR;
    if (p.num >= strlen(s.str))
        return 1;
    str = s.str + p.num;
    if (l.num < 0)
    {
        off = strlen(str) + l.num;
        if (off > 0)
            str += off;
    }
    else
    {
        if (l.num > strlen(str))
            l.num = strlen(str);
        str += l.num;
    }
    strcpy(v->str, str);
    return 1;
}

int sf_strrep(VALUE *v)
{
    int     i, j;
    VALUE   s, n;

    if (!needlpar())            return 0;
    if (!str_expr(&s))          return 0;
    if (!needcomma())           return 0;
    if (!num_expr(&n))          return 0;
    if (!needrpar())            return 0;
    v->typ = T_STR;
    i = 0;
    while (n.num > 0 && i < L_STRING)
    {
        j = 0;
        while (s.str[j])
        {
            if (i < L_STRING)
                v->str[i++] = s.str[j];
            j++;
        }
        n.num--;
    }
    v->str[i] = 0;
    return 1;
}

int sf_strwid(VALUE *v)
{
    VALUE   s;

    if (!needlpar())            return 0;
    if (!str_expr(&s))          return 0;
    if (!needrpar())            return 0;
    v->num = DevGetStrWid(s.str);
    v->typ = T_NUM;
    return 1;
}

int sf_str(VALUE *v)
{
    VALUE   s;
    int     val, len, dec = 0, n, f = 0;
    char    buf[20], *p;

    if (!needlpar())            return 0;
    if (!num_expr(&s))          return 0;
    val = s.num;
    if (!needcomma())           return 0;
    if (!num_expr(&s))          return 0;
    len = s.num;
    if (needcomma())
    {
        if (!num_expr(&s))      return 0;
        dec = s.num;
    }
    if (!needrpar())            return 0;
    if (len > 15)
        return error("invalid LEN");
    if (dec >= len)
        return error("invalid DEC");
    if (val < 0) { val = -val; f = 1; }
    n = 0;
    while (n < dec || val > 0)
    {
        buf[n++] = val % 10 + '0';
        val /= 10;
        if (n == dec)
            buf[n++] = '.';
    }
    if (n == 0)
        buf[n++] = '0';
    if (f)
        buf[n++] = '-';
    while (n < len)
        buf[n++] = ' ';
    buf[n] = 0;
    p = v->str;
    while (n > 0)
        *p++ = buf[--n];
    *p = 0;
    v->typ = T_STR;
    return 1;
}

int sf_ltrim(VALUE *v)
{
    VALUE   s;
    char    *p;

    if (!needlpar())            return 0;
    if (!str_expr(&s))          return 0;
    if (!needrpar())            return 0;
    p = s.str;
    while (*p == ' ')
        p++;
    strcpy(v->str, p);
    v->typ = T_STR;
    return 1;    
}

int sf_rtrim(VALUE *v)
{
    int     n;

    if (!needlpar())            return 0;
    if (!str_expr(v))           return 0;
    if (!needrpar())            return 0;
    n = strlen(v->str);
    while (n > 0 && v->str[n-1] == ' ')
        v->str[--n] = 0;
    v->typ = T_STR;
    return 1;
}

int sf_trim(VALUE *v)
{
    VALUE   s;
    char    *p;
    int     n;

    if (!needlpar())            return 0;
    if (!str_expr(&s))          return 0;
    if (!needrpar())            return 0;
    p = s.str;
    while (*p == ' ')
        p++;
    strcpy(v->str, p);
    n = strlen(v->str);
    while (n > 0 && v->str[n-1] == ' ')
        v->str[--n] = 0;
    v->typ = T_STR;
    return 1;    
}

int sf_dots(VALUE *v)
{
    if (!needlpar())            return 0;
    if (!unit_expr(v))          return 0;
    if (!needrpar())            return 0;
    return 1;
}

int sv_rc(VALUE *v)
{
    v->typ = T_NUM;
    v->num = Rc.num;
    return 1;
}

int sf_eod(VALUE *v)
{
    VALUE   db;

    if (!needlpar())            return 0;
    if (!num_expr(&db))         return 0;
    if (!needrpar())            return 0;
    v->typ = T_NUM;
    v->num = tdb_eof(db.num);
    return 1;
}

int sf_dbflds(VALUE *v)
{
    VALUE   db;

    if (!needlpar())            return 0;
    if (!num_expr(&db))         return 0;
    if (!needrpar())            return 0;
    v->typ = T_NUM;
    v->num = tdb_fldcnt(db.num);
    return 1;
}

int sf_dbfld(VALUE *v)
{
    VALUE   db, fn;
    char    *p;

    if (!needlpar())            return 0;
    if (!num_expr(&db))         return 0;
    if (!needcomma())           return 0;
    if (!num_expr(&fn))         return 0;
    if (!needrpar())            return 0;
    v->typ = T_STR;
    p = tdb_getfld(db.num, fn.num);
    if (p == NULL)
        v->str[0] = 0;
    else
        strcpy(v->str, p);
    return 1;
}

int sf_isnum(VALUE *v)
{
    int     rc;
    VALUE   sym;

    if (!needlpar())            return 0;
    rc = symbol(&sym);
    if (!needrpar())            return 0;
    v->typ = T_NUM;
    if (rc && (sym.typ == T_NUM))
        v->num = 1;
    else
        v->num = 0;
    return 1;
}

int sf_isstr(VALUE *v)
{
    int     rc;
    VALUE   sym;

    if (!needlpar())            return 0;
    rc = symbol(&sym);
    if (!needrpar())            return 0;
    v->typ = T_NUM;
    if (rc && (sym.typ == T_STR))
        v->num = 1;
    else
        v->num = 0;
    return 1;
}

int sf_isvar(VALUE *v)
{
    VALUE   sym;
    int     rc;

    if (!needlpar())            return 0;
    rc = symbol(&sym);
    if (!needrpar())            return 0;
    v->typ = T_NUM;
    v->num = rc;
    return 1;
}

int sf_odd(VALUE *v)
{
    if (!needlpar())            return 0;
    if (!num_expr(v))           return 0;
    if (!needrpar())            return 0;
    v->num = (v->num % 2) > 0;
    v->typ = T_NUM;
    return 1;
}

int sf_float(VALUE *v)
{
    if (!needlpar())            return 0;
    if (!expr(v))               return 0;
    if (!needrpar())            return 0;
    if (v->typ == T_FLT)
        return 1;
    if (v->typ == T_NUM)
    {
        v->flt = ((double) v->num) * 1.0;
        v->typ = T_FLT;
        return 1;
    }
    return 0;
}

int dollar(VALUE *v)
{
    char    buf[L_NAME+1];

    if (!getname(buf, L_NAME, 0))       return 0;

    if (!strcmp(buf, "baseline"))       return sv_baseline(v);
    if (!strcmp(buf, "batchmode"))      return sv_batchmode(v);
    if (!strcmp(buf, "concat"))         return sf_concat(v);
    if (!strcmp(buf, "count"))          return sv_count(v);
    if (!strcmp(buf, "dbfld"))          return sf_dbfld(v);
    if (!strcmp(buf, "dbflds"))         return sf_dbflds(v);
    if (!strcmp(buf, "dots"))           return sf_dots(v);
    if (!strcmp(buf, "eod"))            return sf_eod(v);
    if (!strcmp(buf, "eps"))            return sv_eps(v);
    if (!strcmp(buf, "float"))          return sf_float(v);
    if (!strcmp(buf, "fontcnt"))        return sv_fontcnt(v);
    if (!strcmp(buf, "fontheight"))     return sv_fontheight(v);
    if (!strcmp(buf, "formcnt"))        return sv_formcnt(v);
    if (!strcmp(buf, "formid"))         return sf_formid(v);
    if (!strcmp(buf, "gendate"))        return sv_gendate(v);
    if (!strcmp(buf, "isnum"))          return sf_isnum(v);
    if (!strcmp(buf, "isstr"))          return sf_isstr(v);
    if (!strcmp(buf, "isvar"))          return sf_isvar(v);
    if (!strcmp(buf, "lblcnt"))         return sf_lblcnt(v);
    if (!strcmp(buf, "lblid"))          return sf_lblid(v);
    if (!strcmp(buf, "lblxcnt"))        return sf_lblxcnt(v);
    if (!strcmp(buf, "lblycnt"))        return sf_lblycnt(v);
    if (!strcmp(buf, "ltrim"))          return sf_ltrim(v);
    if (!strcmp(buf, "max"))            return sf_max(v);
    if (!strcmp(buf, "min"))            return sf_min(v);
    if (!strcmp(buf, "now"))            return sv_now(v);
    if (!strcmp(buf, "num"))            return sf_num(v);
    if (!strcmp(buf, "odd"))            return sf_odd(v);
    if (!strcmp(buf, "param"))          return sf_param(v);
    if (!strcmp(buf, "parcnt"))         return sv_parcnt(v);
    if (!strcmp(buf, "propfont"))       return sv_propfont(v);
    if (!strcmp(buf, "rc"))             return sv_rc(v);
    if (!strcmp(buf, "rtrim"))          return sf_rtrim(v);
    if (!strcmp(buf, "str"))            return sf_str(v);
    if (!strcmp(buf, "strftime"))       return sf_strftime(v);
    if (!strcmp(buf, "strlen"))         return sf_strlen(v);
    if (!strcmp(buf, "strpad"))         return sf_strpad(v);
    if (!strcmp(buf, "strrep"))         return sf_strrep(v);
    if (!strcmp(buf, "strseg"))         return sf_strseg(v);
    if (!strcmp(buf, "strstr"))         return sf_strstr(v);
    if (!strcmp(buf, "strtok"))         return sf_strtok(v);
    if (!strcmp(buf, "strwid"))         return sf_strwid(v);
    if (!strcmp(buf, "trim"))           return sf_trim(v);
    if (!strcmp(buf, "version"))        return sv_version(v);
    if (!strcmp(buf, "xbase"))          return sv_xbase(v);
    if (!strcmp(buf, "xsize"))          return sv_xsize(v);
    if (!strcmp(buf, "xpos"))           return sv_xpos(v);
    if (!strcmp(buf, "ybase"))          return sv_ybase(v);
    if (!strcmp(buf, "ysize"))          return sv_ysize(v);
    if (!strcmp(buf, "ypos"))           return sv_ypos(v);
    return 0;
}

int factor(VALUE *v)
{
    int     i;

    skipgap();
    switch (*Cp)
    {
        case '#':
            Cp++;
            if (!factor(v))
                return 0;
            return to_num(v);
        case '!':
            Cp++;
            if (!expr(v))
                return 0;
            if (v->typ != T_NUM)
                return 0;
            v->num = !v->num;
            return 1;
        case '-':
            Cp++;
            if (!expr(v))
                return 0;
            if (v->typ == T_NUM) {
                v->num = -v->num;
                return 1;
            }
            if (v->typ == T_FLT) {
                v->flt = -v->flt;
                return 1;
            }
            return 0;
        case '$':
            Cp++;
            if (!dollar(v))
                return 0;
            return 1;
        case 0x22:
            i = 0;
            Cp++;
            while (*Cp != 0x22)
            {
                if (*Cp == 0)
                    return 0;
                if (i < L_STRING)
                    v->str[i++] = *Cp;
                Cp++;
            }
            Cp++;
            v->str[i] = 0;
            v->typ = T_STR;
            break;
        case '(':
            Cp++;
            if (!expr(v))
                return 0;
            skipgap();
            if (*Cp != ')')
                return 0;
            Cp++;
            break;
        default:
            if (isdigit(*Cp))
                return number(v);
            if (isalpha(*Cp))
                return symbol(v);
            return 0;
    }
    return 1;
}

static int get_compop(void)
{
    if (*Cp == '=')
                { Cp += 1; return CMP_EQ; }
    if (*Cp == '>')
    {
        if (*(Cp+1) == '=')
                { Cp += 2; return CMP_GE; }
        else
                { Cp += 1; return CMP_GT; }
    }
    if (*Cp == '<')
    {
        if (*(Cp+1) == '>')
                { Cp += 2; return CMP_NE; }
        else
        {
            if (*(Cp+1) == '=')
                { Cp += 2; return CMP_LE; }
            else
                { Cp += 1; return CMP_LT; }
        }
    }
    return 0;
}

static int get_logop(void)
{
    if (needchar('&'))
    {
        if (needchar('&'))
            return LOP_AND;
        return BOP_AND;
    }
    if (needchar('|'))
    {
        if (needchar('|'))
            return LOP_OR;
        return BOP_OR;
    }
    return 0;
}

static int mul_val(VALUE *v, VALUE *x)
{
    if (v->typ == T_NUM && x->typ == T_NUM) {
        v->num *= x->num;
        return 1;
    }
    if (v->typ == T_FLT && x->typ == T_FLT) {
        v->flt *= x->flt;
        return 1;
    }
    return 0;
}

static int div_val(VALUE *v, VALUE *x)
{
    if (v->typ == T_NUM && x->typ == T_NUM) {
        v->num /= x->num;
        return 1;
    }
    if (v->typ == T_FLT && x->typ == T_FLT) {
        v->flt /= x->flt;
        return 1;
    }
    return 0;
}

static int mod_val(VALUE *v, VALUE *x)
{
    if (v->typ == T_NUM && x->typ == T_NUM) {
        v->num %= x->num;
        return 1;
    }
    return 0;
}

static int add_val(VALUE *v, VALUE *x)
{
    if (v->typ == T_NUM && x->typ == T_NUM) {
        v->num += x->num;
        return 1;
    }
    if (v->typ == T_FLT && x->typ == T_FLT) {
        v->flt += x->flt;
        return 1;
    }
    return 0;
}

static int sub_val(VALUE *v, VALUE *x)
{
    if (v->typ == T_NUM && x->typ == T_NUM) {
        v->num -= x->num;
        return 1;
    }
    if (v->typ == T_FLT && x->typ == T_FLT) {
        v->flt -= x->flt;
        return 1;
    }
    return 0;
}

static int comp_val_num(int x, int op, int y)
{
    switch (op)
    {
        case CMP_EQ:    return (x == y);
        case CMP_NE:    return (x != y);
        case CMP_LT:    return (x <  y);
        case CMP_LE:    return (x <= y);
        case CMP_GT:    return (x >  y);
        case CMP_GE:    return (x >= y);
        default:        return -1;
    }
}

static int comp_val_flt(double x, int op, double y)
{
    switch (op)
    {
        case CMP_EQ:    return (x == y);
        case CMP_NE:    return (x != y);
        case CMP_LT:    return (x <  y);
        case CMP_LE:    return (x <= y);
        case CMP_GT:    return (x >  y);
        case CMP_GE:    return (x >= y);
        default:        return -1;
    }
}

static int comp_val_str(char *x, int op, char *y)
{
    switch (op)
    {
        case CMP_EQ:    return (strcmp(x, y) == 0);
        case CMP_NE:    return (strcmp(x, y) != 0);
        case CMP_LT:    return (strcmp(x, y) <  0);
        case CMP_LE:    return (strcmp(x, y) <= 0);
        case CMP_GT:    return (strcmp(x, y) >  0);
        case CMP_GE:    return (strcmp(x, y) >= 0);
        default:        return -1;
    }
}

static int comp_val(VALUE *v, int op, VALUE *x)
{
    if (v->typ != x->typ)
        return 0;
    switch (v->typ)
    {
        case T_NUM:
            v->num = comp_val_num(v->num, op, x->num);
            break;
        case T_FLT:
            v->num = comp_val_flt(v->flt, op, x->flt);
            break;
        case T_STR:
            v->num = comp_val_str(v->str, op, x->str);
            v->typ = T_NUM;
            break;
        default:
            return 0;
    }
    return 1;
}

static int log_val(VALUE *v, int op, VALUE *x)
{
    if (v->typ != T_NUM || x->typ != T_NUM)
        return 0;
    switch (op)
    {
        case LOP_AND:
            v->num = v->num && x->num;
            break;
        case LOP_OR:
            v->num = v->num || x->num;
            break;
        case BOP_AND:
            v->num = v->num & x->num;
            break;
        case BOP_OR:
            v->num = v->num | x->num;
            break;
        default:
            return 0;
    }
    return 1;
}

static int term(VALUE *v)
{
    VALUE   x;

    if (!factor(v))
        return 0;
    for (;;)
    {
        skipgap();
        switch (*Cp)
        {
            case '*':
                Cp++;
                if (!factor(&x))
                    return 0;
                if (!mul_val(v, &x))
                    return 0;
                break;
            case '/':
                Cp++;
                if (!factor(&x))
                    return 0;
                if (!div_val(v, &x))
                    return 0;
                break;
            case '%':
                Cp++;
                if (!factor(&x))
                    return 0;
                if (!mod_val(v, &x))
                    return 0;
                break;
            default:
                return 1;
        }
    }
    return 1;
}

int expr1(VALUE *v)
{
    VALUE   x;

    if (!term(v))
        return 0;
    for (;;)
    {
        skipgap();
        switch (*Cp)
        {
            case '+':
                Cp++;
                if (!term(&x))
                    return 0;
                if (!add_val(v, &x))
                    return 0;
                break;
            case '-':
                Cp++;
                if (!term(&x))
                    return 0;
                if (!sub_val(v, &x))
                    return 0;
                break;
            default:
                return 1;
        }
    }
    return 1;
}

int expr2(VALUE *v)
{
    VALUE   x;
    int     op;

    if (!expr1(v))
        return 0;
    skipgap();
    op = get_compop();
    if (op > 0)
    {
        if (!expr1(&x))
            return 0;
        if (!comp_val(v, op, &x))
            return 0;
    }
    return 1;
}

int expr(VALUE *v)
{
    VALUE   x;
    int     op;

    if (!expr2(v))
        return 0;
    skipgap();
    op = get_logop();
    if (op > 0)
    {
        if (!expr2(&x))
            return 0;
        if (!log_val(v, op, &x))
            return 0;
    }
    return 1;
}

int num_expr(VALUE *v)
{
    if (!expr(v))
        return 0;
    if (v->typ != T_NUM)
        return 0;
    return 1;
}

int flt_expr(VALUE *v)
{
    if (!expr(v))
        return 0;
    if (v->typ != T_FLT)
        return 0;
    return 1;
}

int unit_expr(VALUE *v)
{
    double  unit = 1.0;

    if (!expr(v))
        return 0;
    if (v->typ != T_NUM && v->typ != T_FLT)
        return 0;
    if (v->typ == T_NUM)
        v->flt = (double) v->num;
    if (needunit(&unit))
        v->flt = v->flt * unit;
    v->num = (int) (v->flt + 0.5);
    v->typ = T_NUM;
    return 1;
}

int str_expr(VALUE *v)
{
    if (!expr(v))
        return 0;
    if (v->typ != T_STR)
        return 0;
    return 1;
}

int assign(SYMBOL *s, VALUE *v)
{
    s->val.typ = v->typ;
    switch (v->typ)
    {
        case T_NUM:
            s->val.num = v->num;
            break;
        case T_FLT:
            s->val.flt = v->flt;
            break;
        case T_STR:
            strcpy(s->val.str, v->str);
            break;
        default:
            break;
    }
    return 1;
}

