/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 *      fmd.h --    Font Metrics Description
 *
 *          jrs     30.05.98
 */

#ifndef fmd_h
#define fmd_h

typedef struct
{
    unsigned char   first;
    unsigned char   second;
    short           shift;
} KERNPAIR;

typedef struct
{
    
typedef struct      /* Geraetebezogene Daten                            */
{
    char    id[L_ID+1];         /* Device-ID                */
    int     typ;                /* Device-Type              */
    char    name[L_NAME+1];     /* Device-Name              */
    int     xadj;               /* Korrektur X              */
    int     yadj;               /* Korrektur Y              */
} tDEV;

typedef struct      /* Seitenbeschreibung                               */
{
    char    id[L_ID+1];
    int     xlim;
    int     ylim;
    int     lm;
    int     rm;
    int     tm;
    int     bm;
} tPAGE;

typedef struct      /* Informationen ueber den Zeichensatz              */
{
    char    id[L_ID+1];         /* Font-ID                      */
    char    fontname[L_NAME+1]; /* internal fontname            */
    int     points;             /* Punktgroesse (-1 fuer 'alle')*/
    int     refnr;              /* Referenznummer, wenn > 0     */
    int     p_flag;             /* flag: proportional font      */
    int     relwidth;           /* char-width in 1/1000 height  */
    int     attr;               /* font attributes              */
    int     widtab[256];        /* width-table [0..255]         */
} tPSFONT;

typedef struct      /* Parameter des aktuellen Zeichensatzes            */
{
    char    id[L_ID+1];         /* aktueller Fontname                   */
    int     points;             /* aktuelle Punktgre                  */
    int     prop;               /* Flag: proportional oder fixed?       */
    int     y_dots;             /* aktuelle Hoehe in DOTs               */
    int     x_dots;             /* aktuelle Standardbreite in DOTs      */
    int     widtab[256];        /* aktuelle Breitentab. in DOTs         */
} tFONT;


int     DevReset(void);
int     DevConfig(char *filnam);
tPAGE  *DevSelPage(char *id);
int     DevSelFont(tSTYLE *style);
int     DevGetFontXdots(void);
int     DevGetFontYdots(void);
int     DevBeginJob(FILE *ofp);
int     DevEndJob(void);
int     DevBeginPage(void);
int     DevEndPage(void);
int     DevSetWindow(int xpos, int ypos, int xsiz, int ysiz);
int     DevClearWindow(void);
int     DevExit(void);
int     DevSetXY(int x, int y);
int     DevSetX(int x);
int     DevAddY(int n);
int     DevSetAttr(int attr);
int     DevGetChrWid(char c);
int     DevGetStrWid(char *str);
int     DevPrtStr(char *str);
int     DevDrawBox(int xsize, int ysize, int filled);
int     DevDrawRline(int x, int y);
int     DevDrawCircle(int r, int filled);
int     DevSelPen(int pen);
int     DevSelCMYK(tCMYK *color);

#endif	/* devapi_h */
