/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 *  lpu.c -- Das Hauptprogramm
 *
 *      jrs     14.07.96
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>

#include "cfg.h"
#include "expr.h"
#include "lpudefs.h"
#include "utils.h"
#include "devapi.h"
#include "tdbfuncs.h"
#define GLOBAL
#include "lpuvars.h"
#include "lpufuncs.h"

tPAGE  *Page = NULL;
tFORM  *Form = NULL;
tFORM   Forms[MAXFRM];
int     Frmcnt = 0, Frmndx = 0;
tLABEL *pLbl = NULL;
tLABEL  Labels[MAXLBL];
int     Lblcnt = 0, Skipcnt = 0;
int     Attr;

tWIN    Wtab[MAXWIN];
int     Wcnt = 0;

ITEM   *Pens   = NULL;
ITEM   *Inks   = NULL;
ITEM   *Styles = NULL;

FILE    *Ofp, *Cfp;
char    Cmdfile[L_FNAME+1];
char    Outfile[L_FNAME+1];
char    Devfile[L_FNAME+1];
char    Tabfile[L_FNAME+1];
char    Parfile[L_FNAME+1];

int     Txtadj = LEFT;          /* Textausrichtung                          */
char    Prtbuf[256], *Pp = Prtbuf;
char	Tmplin[256];
char	Wrklin[256];
char	Watchstr[256];
int	Watchflg = FALSE;

int     Exitflg = FALSE;

int     Pen = 10;
tCMYK   Color = { 0, 0, 0, 100 };
int     Dash = 0;

void option_step(int flg);

static void sighandler(int signo)
{
    if (signo == SIGINT)
        option_step(TRUE);
}

int get_position(int *x, int *y)
{
    skipgap();
    if (!needlpar())
        return error("'(' expected");
    skipgap();
    if (!getunit(x))
        return error("x-pos expected");
    skipgap();
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(y))
        return error("y-pos expected");
    skipgap();
    if (!needrpar())
        return error("')' expected");
    return TRUE;
}

int get_size(tSIZE *size)
{
    skipgap();
    if (!needlpar())
        return error("'(' expected");
    skipgap();
    if (!getunit(&(size->xsiz)))
        return error("x-size expected");
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&(size->ysiz)))
        return error("y-size expected");
    skipgap();
    if (!needrpar())
        return error("')' expected");
    return TRUE;
}

/* ----- Funktionen fr die ITEM-Verwaltung ----- */

static ITEM *fnd_item(ITEM *list, char *id)
{
    while (list != NULL)
    {
        if (!strcmp(id, list->id))
            return list;
        list = list->next;
    }
    return NULL;
}


int get_pen(int *width)
{
    char    id[L_ID+1];
    ITEM    *p;

    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!getname(id, L_ID, FALSE))
            return error("pen-id expected");
        p = fnd_item(Pens, id);
        if (p == NULL)
            return error("undefined pen-id");
        *width = p->pen;
        return TRUE;
    }
    if (!getunit(width))
        return error("invalid pen value");
    return TRUE;
}

int get_ink(tCMYK *ink)
{
    int     *val;
    char    id[L_ID+1];
    ITEM    *p;

    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!getname(id, L_ID, FALSE))
            return error("ink-id expected");
        p = fnd_item(Inks, id);
        if (p == 0)
            return error("unknown ink-id");
        memcpy(ink, &(p->ink), sizeof(tCMYK));
        return TRUE;
    }
    ink->cyan    = 0;
    ink->magenta = 0;
    ink->yellow  = 0;
    ink->black   = 0;
    if (*Cp != '(')
        return error("'(' expected");
    Cp++;
    skipgap();
    while (isalpha(*Cp))
    {
        switch (*Cp)
        {
            case 'c':   val = &(ink->cyan);     break;
            case 'm':   val = &(ink->magenta);  break;
            case 'y':   val = &(ink->yellow);   break;
            case 'k':   val = &(ink->black);    break;
            default:    return error("unknown ink component");
        }
        Cp++;
        if (!getdec(val))
            return error("missing ink value");
        skipgap();
    }
    if (*Cp != ')')
        return error("')' expected");
    Cp++;
    return TRUE;
}

int get_style(tSTYLE *style)
{
    char    id[L_ID+1];
    ITEM    *p;

    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!getname(id, L_ID, FALSE))
            return error("style-id expected");
        p = fnd_item(Styles, id);
        if (p == 0)
            return error("unknown style-id");
        memcpy(style, &(p->style), sizeof(tSTYLE));
        return TRUE;
    }
    if (!getname(style->family, L_ID, FALSE))
        return error("family expected");
    skipgap();
    if (!getdec(&(style->points)))
        return error("points expected");
    skipgap();
    style->attr = 0;
    while (isalpha(*Cp))
    {
        switch (*Cp)
        {
            case 'B':       style->attr |= AF_BOLD;     break;
            case 'I':       style->attr |= AF_ITALIC;   break;
            default:        return error("invalid attribute");
        }
        Cp++;
    }
    return TRUE;
}


/* ----- Optionen setzen ----- */

void option_step(int flg)
{
    OptStep = flg;
    if (pCtx != NULL)
        pCtx->step = flg;
}

void option_trace(int flg)
{
    OptTrace = flg;
    if (pCtx != NULL)
        pCtx->trace = flg;
}

/* ----- diverse Funktionen ----- */

void strip_lf(char buf[])
{
    int     len = strlen(buf);

    if (len > 0 && buf[len-1] == '\n')
        buf[--len] = 0;
}

void sel_dash(int dash)
{
    if (dash != Dash)
    {
        DevSelDash(dash);
        Dash = dash;
    }
}

int sel_form(char *name)
{
    int     i;

    i = 0;
    while (i < MAXFRM)
    {
        if (!strcmp(name,Forms[i].id))
        {
            Form = &Forms[i];
            Page = DevSelPage(Form->page);
            if (Page == NULL)
                return FALSE;
            Frmndx = i;
            return TRUE;
        }
        i++;
    }
    return FALSE;
}

static tLABEL *find_lblid(char *name)
{
    int     i;

    for (i=0; i<Form->lblcnt; i++)
        if (!strcmp(name, Labels[Form->lbl1st+i].id))
            return &Labels[Form->lbl1st+i];
    return NULL;
}

#if 0
int warning(char *fmt, ...)
{
    va_list arglst;
    char    buf[256];

    va_start(arglst, fmt);
    vsprintf(buf, fmt, arglst);
    va_end(arglst);
    fprintf(stderr,"warning: %s\n",buf);
    if (Ctxcnt > 0)
    {
        fprintf(stderr, "  in proc '%s'\n", pCtx->proc->procname);
        fprintf(stderr, "    txt='%s'\n", Tmplin);
    }
    return TRUE;
}

int error(char *fmt, ...)
{
    va_list arglst;
    char    buf[256];

    va_start(arglst, fmt);
    vsprintf(buf, fmt, arglst);
    va_end(arglst);
    fprintf(stderr,"error: %s\n",buf);
    if (Ctxcnt > 0)
    {
        fprintf(stderr, "  in proc '%s'\n", pCtx->proc->procname);
        fprintf(stderr, "    txt='%s'\n", Tmplin);
    }
    return FALSE;
}

void verbose(int level, char *fmt, ...)
{
    va_list arglst;

    if (OptVerbose >= level)
    {
        va_start(arglst, fmt);
        vfprintf(stderr, fmt, arglst);
        va_end(arglst);
    }
}
#endif

void prt_line(int adj)
{
    int     len, xoff, gap;

    *Pp = 0;
    len = DevGetStrWid(Prtbuf);
    if (len > 0)
    {
        xoff = 0;
        gap = W->xsiz - W->lm - W->rm - len;
        if (gap > 0)
            switch (adj) {
                case LEFT:      xoff = 0;       break;
                case RIGHT:     xoff = gap;     break;
                case CENTER:    xoff = gap / 2; break;
            }
        DevSetXY(W->lm + W->xref + xoff, W->yref + W->ypos);
        DevPrtStr(Prtbuf, 0);
    }
    Pp = Prtbuf;
}

int setborder(tWIN *win, int tm, int bm, int lm, int rm)
{
    verbose(1, "setborder: tm=%d, bm=%d, lm=%d, rm=%d\n", tm, bm, lm, rm);
    win->tm   = tm;
    win->bm   = bm;
    win->lm   = lm;
    win->rm   = rm;
    win->xpos = lm;
    win->ypos = tm;
    win->xref = 0;
    win->yref = 0;
    return TRUE;
}

int setwindow(tWIN *win, int x, int y, int w, int h, int f)
{
    int     xabs, yabs;

    verbose(1, "setwindow: x=%d, y=%d, w=%d, h=%d, f=%d\n", x, y, w, h, f);
    win->xoff  = x;
    win->yoff  = y;
    win->xsiz  = w;
    win->ysiz  = h;
    win->xpos  = 0;
    win->ypos  = 0;
    setborder(win, 0, 0, 0, 0);
    xabs  = pLbl->xbase + (pLbl->xsize + pLbl->xskip) * pLbl->xlab;
    yabs  = pLbl->ybase + (pLbl->ysize + pLbl->yskip) * pLbl->ylab;
    DevSetWindow(xabs + win->xoff, yabs + win->yoff, win->xsiz, win->ysiz, f);
    verbose(1, "setwindow: Lblndx=%d, Lblid='%s', Xlab=%d, Ylab=%d\n",
        pLbl->ndx, pLbl->id, pLbl->xlab, pLbl->ylab);
    W = win;
    return TRUE;
}

int setlabel(void)
{
    Wcnt = 0;
    pLbl->xlab  = pLbl->ndx % pLbl->xcnt;
    pLbl->ylab  = pLbl->ndx / pLbl->xcnt;
    setwindow(&Wtab[Wcnt], 0, 0, pLbl->xsize, pLbl->ysize, 0);
    return TRUE;
}

void put_chr(char c)
{
    *Pp++ = c;
}

int prt_hline(int len, int pen, int adj)
{
    int     yp, gap, xoff;

    yp = W->ypos;
    gap = W->xsiz - len;
    if (gap > 0)
        switch (adj) {
            case LEFT:      xoff = 0;       break;
            case RIGHT:     xoff = gap;     break;
            case CENTER:    xoff = gap / 2; break;
        }
    DevSetXY(W->xref /* + W->xpos*/ + xoff, W->yref + W->ypos);
    DevSelPen(pen);
    DevDrawRline(len, 0);
    W->xpos = W->lm; // was: 0
    W->ypos = yp + DevGetFontYdots();
    DevSetXY(W->xref + W->xpos, W->yref + W->ypos);
    DevSelPen(Pen);     /* restore current default pen */
    return TRUE;
}

/*** textuelle Ersetzung einer Zeile ***/
/*
**  folgende Ersetzungen werden durchgefuehrt:
**      Token       erzeugt
**      ----------------------------------------------------------
**      \\          ein Backslash
**
**      \0          Name des aktuellen Moduls
**      \n          aktueller Parameter n (n='1'..'9')
**      \#          Anzahl der aktuellen Parameter
**      \?          Status des letzten Moduls
**      \$          Wiederholungszaehler fuer das aktuelle Modul
**
**      \@0         reserviert
**      \@n         aktuelles Datenbankfeld (n='1'..'9')
**      \@#         Anzahl der Datenfelder im aktuellen Satz
**      \@$         Nummer des aktuellen Datensatzes (1..n, 0=EOF)
**      \@?         EOF-Status der Datenbank
**
**      \(id)       Symbol 'id', ggf. nach Text konvertiert
**      \{expr}     Ausdruck 'expr', ggf. nach Text konvertiert
**      \[env]      Environment-Variable 'env'
*/

static char *Xp;
static int   Xn, Xi;
static void apnd_chr(char c)
{
    if (Xi >= Xn)
        return;
    *Xp++ = c;
    Xi++;
}

static void apnd_str(char *s)
{
    while (*s)
        apnd_chr(*s++);
}

int convert_line(char *p, int siz)
{
    char    buf[128], *bp;
    VALUE   v;

    Xp = p;
    Xn = siz;
    Xi = 0;
    while (*Cp)
    {
        if (*Cp != '\\')
        {
            apnd_chr(*Cp++);
            continue;
        }
        Cp++;
        switch (*Cp)
        {
            case '\\':
                apnd_chr(*Cp++);
                break;
            case '(':
                Cp++;
                if (!symbol(&v))
                    return error("cvt: error in symbol");
                if (*Cp != ')')
                    return error("cvt: ')' expected");
                if (v.typ == T_NUM)
                    sprintf(v.str, "%d", v.num);
                apnd_str(v.str);
                Cp++;
                break;
            case '{':
                Cp++;
                if (!expr(&v))
                    return error("cvt: error in expression");
                if (*Cp != '}')
                    return error("cvt: '}' expected");
                if (v.typ == T_NUM)
                    sprintf(v.str, "%d", v.num);
                apnd_str(v.str);
                Cp++;
                break;
            case '[':
                Cp++;
                if (!getname(buf, 64, FALSE))
                    return error("env-symbol expected");
                if (*Cp != ']')
                    return error("cvt: ']' expected");
                Cp++;
                v.typ = T_STR;
                v.str[0] = 0;
                bp = getenv(buf);
                if (bp != NULL)
                    strcpy(v.str, bp);
                apnd_str(v.str);
                break;
            case '@':
                Cp++;
                if (*Cp == '#')
                {
                    Cp++;
                    sprintf(buf, "%d", tdb_fldcnt(0));
                    apnd_str(buf);
                    break;
                }
                if (*Cp >= '1' && *Cp <= '9')
                {
                    p = tdb_getfld(0, *Cp++ - '0');
                    if (p == NULL)
                        break;
                    apnd_str(p);
                    break;
                }
                if (*Cp == '$')
                {
                    Cp++;
                    sprintf(buf, "%d", tdb_recno(0));
                    apnd_str(buf);
                    break;
                }
                if (*Cp == '?')
                {
                    Cp++;
                    sprintf(buf, "%d", tdb_eof(0));
                    apnd_str(buf);
                    break;
                }
                put_chr(*Cp++);
                break;
            case '?':
                if (Rc.typ == T_NUM)
                    sprintf(Rc.str, "%d", v.num);
                apnd_str(Rc.str);
                Cp++;
                break;
            default:
                apnd_chr(*Cp++);
                break;
        }
    }
    *Xp = 0;
    return TRUE;
}

int do_trace(void)
{
    char    *s;

    fprintf(stderr, "[%d]: %s\n", Ctxcnt, Cp);
    if (Watchflg)
    {
        s = Cp;
        fprintf(stderr, "Watch: ");
        Cp = Watchstr;
        cmd_debug();
        Cp = s;
    }
    return 1;
}

static void disp_stack(void)
{
    int     n = Ctxcnt;

    while (n >= 0)
    {
        fprintf(stderr, "%4d: %s\n", n, Ctxtab[n].proc->procname);
        n--;
    }
}

static void disp_val(VALUE *v)
{
    switch (v->typ)
    {
        case T_NUM:
            fprintf(stderr, "#%ld\n", v->num);
            break;
        case T_STR:
            fprintf(stderr, "\"%s\"\n", v->str);
            break;
        case T_FLT:
            fprintf(stderr, "#%lf\n", v->flt);
            break;
        default:
            fprintf(stderr, "?type\n");
            break;
    }
}

int do_step(void)
{
    char    cmd[256], *sp;
    VALUE   v;

    sp = Cp;
    for (;;)
    {
        fprintf(stderr, "--> ");
        fgets(cmd, 255, stdin);
        strip_lf(cmd);
        Cp = cmd;
        skipgap();
        switch (*Cp)
        {
            case 0:
                Cp = sp;
                return TRUE;
            case '?':
                fprintf(stderr, "  ?           help message\n");
                fprintf(stderr, "  aVAR=EXPR   assign EXPR to VAR\n");
                fprintf(stderr, "  c           display callstack\n");
                fprintf(stderr, "  eEXPR       evaluate and display EXPR\n");
                fprintf(stderr, "  g           go\n");
                fprintf(stderr, "  q           quit\n");
                fprintf(stderr, "  w[EXPR ...] watch list of expressions\n");
                fprintf(stderr, "  CR          single step\n");
                break;
            case 'g':
                option_step(FALSE);
                Cp = sp;
                return TRUE;
            case 'c':
                disp_stack();
                break;
            case 'a':
                Cp++;
                cmd_let();
                break;
            case 'e':
                Cp++;
                if (expr(&v))
                    disp_val(&v);
                break;
            case 'q':
                Exitflg = TRUE;
                Cp = sp;
                return TRUE;
            case 'w':
                Watchflg = FALSE;
                Cp += 1;
                skipgap();
                if (*Cp > ' ')
                {
                    strcpy(Watchstr, Cp);
                    Watchflg = TRUE;
                }
                break;
            default:
                fprintf(stderr, "???\n");
                break;
        }
    }
}

int do_txt(void)
{
    char    *p;

    p = Tmplin;
    p++;
    while (*p)
        put_chr(*p++);
    return TRUE;
}

int do_nl(int cnt)
{
    prt_line(Txtadj);
    W->xpos  = W->lm; // was: 0
    while ( cnt-- )
        W->ypos += DevGetFontYdots();
    return TRUE;
}

int do_base(int x, int y)
{
    W->xref = x;
    W->yref = y;
    return TRUE;
}

int do_setxy(int x, int y)
{
    // @@@ pruefen!
    W->xpos = x;
    W->ypos = y;
    return TRUE;
}



int cmd_defink(VOID)
{
    char    id[L_ID+1];
    tCMYK   ink;
    ITEM    *item;

    skipgap();
    if (!getname(id, L_ID, FALSE))
        return error("ink-id expected");
    skipgap();
    if (!get_ink(&ink))
        return FALSE;
    item = (ITEM *) calloc(1, sizeof(ITEM));
    if (item == NULL)
        return error("out of memory");
    strcpy(item->id, id);
    memcpy(&(item->ink), &ink, sizeof(tCMYK));
    item->next = Inks;
    Inks = item;
    return TRUE;
}

int cmd_defpen(VOID)
{
    char    id[L_ID+1];
    int     pen;
    ITEM    *item;

    skipgap();
    if (!getname(id, L_ID, FALSE))
        return error("pen-id expected");
    skipgap();
    if (!get_pen(&pen))
        return FALSE;
    item = (ITEM *) calloc(1, sizeof(ITEM));
    if (item == NULL)
        return error("out of memory");
    strcpy(item->id, id);
    item->pen = pen;
    item->next = Pens;
    Pens = item;
    return TRUE;
}

int cmd_defstyle(VOID)
{
    char    id[L_ID+1];
    tSTYLE  style;
    ITEM    *item;

    skipgap();
    if (!getname(id, L_ID, FALSE))
        return error("style-id expected");
    skipgap();
    if (!get_style(&style))
        return FALSE;
    item = calloc(1, sizeof(ITEM));
    if (item = NULL)
        return error("out of memory");
    strcpy(item->id, id);
    memcpy(&(item->style), &style, sizeof(tSTYLE));
    item->next = Styles;
    Styles = item;
    return TRUE;
}

int do_db_use(void)
{
    int     n;

    skipgap();
    if (!getdec(&n))
        return error("area expected");
    if (n < 1)
        return error("invalid area");
    return tdb_use(n);
}

int do_db_open(void)
{
    int     n, rc;
    char    dbnam[L_FNAME+1], sep;

    n = 0;
    sep = 0;
    skipgap();
    if (*Cp == '#')
    {
        Cp++;
        if (!getdec(&n))
            return error("area expected");
        if (n < 1)
            return error("invalid area");
        skipgap();
    }
    if (*Cp == '/')
    {
        Cp++;
        if (*Cp)
            sep = *Cp++;
        skipgap();
    }
    if (!getfname(dbnam, L_FNAME-4))
        return error("database-name expected");
    rc = tdb_open(n, dbnam, 0, 0, sep);
    if (!rc)
        return rc;
    return tdb_top();
}

int do_db_close(void)
{
    return tdb_close();
}

int do_db_top(void)
{
    return tdb_top();
}

int do_db_next(void)
{
    return tdb_next();
}

int do_db_find(void)
{
    int     n;

    skipgap();
    if (!getdec(&n))
        return error("fieldnumber expected");
    skipgap();
    if (*Cp == 0)
        return error("pattern expected");
    return tdb_find(n, Cp);
}



int cmd_db(void)
{
    char    cmd[20];

    skipgap();
    if (!getname(cmd, 19, FALSE))
        return error("keyword expected");
    if (!strcmp(cmd, "use"))            return do_db_use();
    if (!strcmp(cmd, "open"))           return do_db_open();
    if (!strcmp(cmd, "close"))          return do_db_close();
    if (!strcmp(cmd, "top"))            return do_db_top();
    if (!strcmp(cmd, "next"))           return do_db_next();
    if (!strcmp(cmd, "find"))           return do_db_find();
    return error("unknown keyword");
}

int cmd_form(void)
{
    int     i;

    if (Form != NULL && pLbl != NULL)
        return error("form already defined");
    skipgap();
    if (!sel_form(Cp))
        return error("unknown form");
    for (i=0; i<Form->lblcnt; i++)
        Labels[Form->lbl1st+i].ndx = -1;
    pLbl = NULL;
    return TRUE;
}

int cmd_label(void)
{
    int     new = TRUE;
    char    ident[L_ID+1];

    if (Form == NULL)
        return error("no form selected");
    if (pLbl == NULL)
        pLbl = &Labels[Form->lbl1st];
    else
        new = FALSE;
    skipgap();
    if (*Cp)
    {
        if (!getname(ident, L_ID, FALSE))
            return error("label id expected");
        pLbl = find_lblid(ident);
        if (pLbl == NULL)
            return error("invalid label id");
    }
    if (new)
    {
        if (Skipcnt > 0)
        {
            if (Skipcnt >= pLbl->xcnt * pLbl->ycnt)
                return error("invalid skipcnt");
	    pLbl->ndx = Skipcnt - 1;
            Skipcnt = 0;
        }
        DevBeginPage();
    }

    pLbl->ndx++;
    if (pLbl->ndx >= pLbl->xcnt * pLbl->ycnt)
    {
        DevEndPage(1);
        DevBeginPage();
        pLbl->ndx = 0;
    }

    setlabel();
    return TRUE;
}

int cmd_window(void)
{
    int     x, y, w, h, f = 0;

    skipgap();
    if (*Cp == 0)
    {
        setlabel();
        return TRUE;
    }
    x = 0; y = 0;
    if (*Cp == '@')
    {
        Cp++;
        if (!get_position(&x, &y))
            return FALSE;
    }
    skipgap();

    while (needchar('.'))
    {
        char    opt[21];
        if (!getname(opt, 20, FALSE))
            return error("option expected");
        if (!strcmp(opt, "frame")) {
            f = 1;
        } else {
            return error("unknown option");
        }
        skipgap();
    }

    if (!getunit(&w))
        return error("width expected");
    skipgap();
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&h))
        return error("height expected");

    setwindow(W, x, y, w, h, f);
    return TRUE;
}

int cmd_border(void)
{
    int     lm, rm, tm, bm;

    if (W == NULL)
        return error("no window defined");
    skipgap();
    if (!getunit(&tm))
        return error("top margin expected");
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&bm))
        return error("bottom margin expected");
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&lm))
        return error("left margin expected");
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&rm))
        return error("right margin expected");
    Wcnt++;
    setborder(W, tm, bm, lm, rm);
    return TRUE;
}

int cmd_eject(int show)
{
    int     i;

    if (pLbl != NULL)
        DevEndPage(show);
    pLbl = NULL;
    for (i=0; i<Form->lblcnt; i++)
        Labels[Form->lbl1st+i].ndx = -1;
    return TRUE;
}

// .pen WIDTH
//
int cmd_pen(void)
{
    int     n;

    if (!get_pen(&n))
        return error("linewidth expected");
    Pen = n;
    return TRUE;
}

// .ink [cN] [mN] [yN] [kN]
//
int cmd_ink(void)
{
    tCMYK   color;

    if (!get_ink(&color))
        return FALSE;
    DevSelCMYK(&color);
    memcpy(&Color, &color, sizeof(tCMYK));
    return TRUE;
}

// .base XPOS, YPOS
//
int cmd_base(void)
{
    int     x, y;

    Cp++;
    skipgap();
    if (!getunit(&x))
        return error("x-base expected");
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&y))
        return error("y-base expected");
    return do_base(x, y);
}

// .setxy XPOS, YPOS
//
int cmd_setxy(void)
{
    int     x, y;

    Cp++;
    skipgap();
    if (!getunit(&x))
        return error("x-pos expected");
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&y))
        return error("y-pos expected");
    return do_setxy(x, y);
}

int cmd_font(void)
{
    return error("not implemented");
}

// .style STYLEDEF
//
int cmd_style(void)
{
    tSTYLE  style;

    if (!get_style(&style))
        return FALSE;
    if (DevSelFont(&style) == 0)
        return error("illegal style");
    Par.x_unit = DevGetFontXdots();
    Par.t_unit = DevGetFontXdots();
    Par.y_unit = DevGetFontYdots();
    Par.r_unit = DevGetFontYdots();
    // @@@ Attr = Font.defattr;
    return TRUE;
}

// .nl [LINES]
//
int cmd_nl(void)
{
    int     n;

    if (pLbl == NULL)           return error("no label selected");
    n = 1;
    skipgap();
    if (*Cp)
        if (!getdec(&n))
            return error("linecount expected");
#if 1
    do_nl(n);
#else
    W->xpos = W->lm; // was: 0
    while (n--)
    {
        prt_line(Txtadj);
        W->ypos += DevGetFontYdots();
    }
#endif
    return TRUE;
}

// .[ADJ]text :TEXT
//
int cmd_text(int adj)
{
    skipgap();
    if (!needchar(':'))
        return error("colon expected");
    while (*Cp)
        put_chr(*Cp++);
    prt_line(adj);
    return TRUE;
}

int cmd_print(int uflag)
{
    int     x, y, len, at = FALSE;
    VALUE   v;

    x = W->xpos;
    y = W->ypos;
    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!get_position(&x, &y))
            return FALSE;
        at = TRUE;
    }
    for (;;)
    {
        skipgap();
        if (!str_expr(&v))
            return error("error in string expression");
        len = DevGetStrWid(v.str);
        if (len > 0)
        {
            DevSetXY(W->xref + x, W->yref + y);
            DevPrtStr(v.str, uflag);
            x += len;
        }
        skipgap();
        if (*Cp != ',')
            break;
        Cp++;
    }
    if (!at)
        W->xpos = x;
    return TRUE;
}

int cmd_debug(void)
{
    VALUE   v;

    do
    {
        if (!expr(&v))
            return error("expression error");
        switch (v.typ)
        {
            case T_NUM:
                fprintf(stderr, "%d", v.num);
                break;
            case T_STR:
                fprintf(stderr, "%s", v.str);
                break;
            case T_FLT:
                fprintf(stderr, "%f", v.flt);
            default:
                break;
        }
    } while (needcomma());
    fprintf(stderr, "\n");
    return TRUE;
}

int cmd_stop(void)
{
    if (OptStdin)
        return error(".stop not supported in STDIN mode");

    option_step(TRUE);
    return TRUE;
}

int cmd_cont(void)
{
    option_step(FALSE);
    return TRUE;
}

int cmd_tron(void)
{
    option_trace(TRUE);
    return TRUE;
}

int cmd_troff(void)
{
    option_trace(FALSE);
    return TRUE;
}

int cmd_watch(void)
{
    Watchflg = FALSE;
    skipgap();
    if (*Cp > ' ')
    {
        if (strlen(Cp) > 250)
            return error("watchlist too long");
        strcpy(Watchstr, Cp);
        Watchflg = TRUE;
    }
    return TRUE;
}

int cmd_input(void)
{
    SYMBOL  *sym;
    char    buf[L_NAME+1];

    skipgap();
    if (!getname(buf, L_NAME, FALSE))
        return error("varname expected");
    sym = symFindAll(buf);
    if (sym == NULL)
        return error("unknown variable");
    skipgap();
    if (!needchar(':'))
        return error("colon expected");

    if (OptStdin)
        return error(".input not supported in STDIN mode");

    fprintf(stderr, "%s: ", Cp);
    fgets(sym->val.str, L_STRING, stdin);
    strip_lf(sym->val.str);
    sym->val.typ = T_STR;
    return TRUE;
}

// .hline WIDTH [.pen WIDTH] [.center | .left | .right]
//
int cmd_hline(void)
{
    int     n, pen = Pen, adj = Txtadj;
    char    opt[10];

    if (pLbl == NULL)
        return error("no label defined");
    skipgap();
    if (!getunit(&n))
        return error("width expected");
    skipgap();
    while (*Cp == '.')
    {
        Cp++;
        if (!getname(opt, 9, FALSE))
            error("option expected");
        skipgap();
        if (!strcmp(opt, "pen")) {
            if (!get_pen(&pen))
                return FALSE;
        }
        else if (!strcmp(opt, "center")) {
            adj = CENTER;
        }
        else if (!strcmp(opt, "left")) {
            adj = LEFT;
        }
        else if (!strcmp(opt, "right")) {
            adj = RIGHT;
        }
        else
            return error("invalid option");
    }
    return prt_hline(n, pen, adj);
}

int cmd_frame(void)
{
    sel_dash(0);
    DevSetXY(0, 0);
    DevSelPen(1);
    DevDrawBox(W->xsiz, W->ysiz, FALSE);
    return TRUE;
}

// .gbox [@( XPOS, YPOS)] WIDTH, HEIGHT [.fill] [.pen PENDEF] [.ink INKDEF]
//
int cmd_gbox(void)
{
    int     x, y, w, h;
    int     pen = Pen, filled = FALSE, dash = 0;
    char    opt[10];
    tCMYK   ink;

    x = W->xpos;
    y = W->ypos;
    memcpy(&ink, &Color, sizeof(tCMYK));
    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!get_position(&x, &y))
            return FALSE;
    }
    skipgap();
    if (!getunit(&w))   return error("width expected");
    if (!needcomma())   return error("comma expected");
    skipgap();
    if (!getunit(&h))   return error("heigth expected");
    skipgap();
    while (*Cp == '.')
    {
        Cp++;
        if (!getname(opt, 9, FALSE))
            return error("option expected");
        if (!strcmp(opt, "fill"))
        {
            filled = TRUE;
        }
        else if (!strcmp(opt, "pen"))
        {
            if (!get_pen(&pen))
                return FALSE;
        }
        else if (!strcmp(opt, "ink"))
        {
            if (!get_ink(&ink))
                return FALSE;
        }
        else if (!strcmp(opt, "dashed"))
        {
            dash |= 0x0001;
        }
        else if (!strcmp(opt, "dotted"))
        {
            dash |= 0x0002;
        }
        else
            return error("invalid option");
        skipgap();
    }
    sel_dash(dash);
    DevSetXY(x, y);
    DevSelPen(pen);
    DevSelCMYK(&ink);
    DevDrawBox(w, h, filled);
    return TRUE;
}

// .gline [@(XPOS, YPOS) WIDTH, HEIGHT [.pen PENDEF] [.ink INKDEF]
//        [.dashed] [.dotted]
//
int cmd_gline(void)
{
    int     x, y, w, h;
    int     pen = Pen, dash = 0;
    char    opt[10];
    tCMYK   ink;

    x = W->xpos;
    y = W->ypos;
    memcpy(&ink, &Color, sizeof(tCMYK));
    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!get_position(&x, &y))
            return FALSE;
    }
    skipgap();
    if (!getunit(&w))   return error("width expected");
    if (!needcomma())   return error("comma expected");
    skipgap();
    if (!getunit(&h))   return error("heigth expected");
    skipgap();
    while (*Cp == '.')
    {
        Cp++;
        if (!getname(opt, 9, FALSE))
            return error("option expected");
        if (!strcmp(opt, "pen"))
        {
            if (!get_pen(&pen))
                return FALSE;
        }
        else if (!strcmp(opt, "ink"))
        {
            if (!get_ink(&ink))
                return FALSE;
        }
        else if (!strcmp(opt, "dashed"))
        {
            dash |= 0x0001;
        }
        else if (!strcmp(opt, "dotted"))
        {
            dash |= 0x0002;
        }
        else
            return error("invalid option");
        skipgap();
    }
    sel_dash(dash);
    DevSetXY(x, y);
    DevSelPen(pen);
    DevSelCMYK(&ink);
    DevDrawRline(w, h);
    return TRUE;
}

// .gcircle [@(XPOS, YPOS) RADIAN [.fill] [.pen PENDEF] [.ink INKDEF]
//
int cmd_gcircle(void)
{
    int     x, y, r, filled = FALSE, pen = Pen;
    char    opt[10];
    tCMYK   ink;

    x = W->xpos;
    y = W->ypos;
    memcpy(&ink, &Color, sizeof(tCMYK));
    skipgap();
    if (*Cp == '@')
    {
        Cp++;
        if (!get_position(&x, &y))
            return FALSE;
    }
    skipgap();
    if (!getunit(&r))   return error("radian expected");
    skipgap();
    while (*Cp == '.')
    {
        Cp++;
        if (!getname(opt, 9, FALSE))
            return error("option expected");
        if (!strcmp(opt, "fill"))
        {
            filled = TRUE;
        }
        else if (!strcmp(opt, "pen"))
        {
            if (!get_pen(&pen))
                return FALSE;
        }
        else if (!strcmp(opt, "ink"))
        {
            if (!get_ink(&ink))
                return FALSE;
        }
        else
            return error("invalid option");
        skipgap();
    }
    sel_dash(0);
    DevSetXY(x, y);
    DevSelPen(Pen);
    DevSelCMYK(&ink);
    DevDrawCircle(r, filled);
    return TRUE;
}

int cmd_left(void)
{
    Txtadj = LEFT;
    return TRUE;
}

int cmd_right(void)
{
    Txtadj = RIGHT;
    return TRUE;
}

int cmd_center(void)
{
    Txtadj = CENTER;
    return TRUE;
}

int do_use(int db, char *cmd)
{
    while (tdb_eof(db) == 0)
    {
        Cp = cmd;
        if (!cmd_do())
            return FALSE;
    }
    return TRUE;
}

int cmd_use(void)
{
    char    *sp;
    int     db, rc;

    skipgap();
    if (!getdec(&db))
        return error("area expected");
    if (db < 0)
        return error("invalid area");
    skipgap();
    sp = push_str(Cp);
    if (sp == NULL)
        return error("out of memory");
    rc = do_use(db, sp);
    free(sp);
    return rc;
}

int cmd_var(void)
{
    char    buf[L_NAME+1];

    for (;;)
    {
        skipgap();
        if (!getname(buf, L_NAME, FALSE))
            return error("varname expected");
        //if (symFindAll(buf) != NULL)
        if (symFindLoc(buf) != NULL)
            return error("variable already defined");
        if (symAdd(&(pCtx->vars), buf) == NULL)
            return error("out of memory");
        skipgap();
        if (*Cp == 0 || *Cp == ';')
            break;
        if (*Cp != ',')
            return error("comma expected");
        Cp++;
    }
    return TRUE;
}

int cmd_let(void)
{
    SYMBOL  *sym;
    VALUE   val;
    char    buf[L_NAME+1];

    skipgap();
    if (!getname(buf, L_NAME, FALSE))
        return error("varname expected");
    sym = symFindAll(buf);
    if (sym == NULL)
        return error("unknown variable");
    skipgap();
    if (!needchar('='))
        return error("'=' expected");
    if (!expr(&val))
        return FALSE;
    if (!assign(sym, &val))
        return FALSE;
    return TRUE;
}

int cmd_goto(void)
{
    char    label[20], buf[20];
    tLINE   *lp;

    skipgap();
    if (!getname(label, 19, FALSE))
        return error("label expected");
    lp = pCtx->proc->first;
    while (lp != NULL)
    {
        Cp = lp->line;
        if (*Cp == ':')
        {
            Cp++;
            skipgap();
            if (!getname(buf, 19, FALSE))
                return error("invalid label");
            if (!strcmp(label, buf))
            {
                pCtx->line = lp;
                return TRUE;
            }
        }
        lp = lp->next;
    }
    return error("label not found");
}

int cmd_do(void)
{
    char    name[20];
    tPROC   *pp;

    skipgap();
    if (!getname(name, 19, FALSE))
        return error("label expected");
    pp = Main.next;
    while (pp != NULL)
    {
        if (!strcmp(name, pp->procname))
        {
            OptStep  = pCtx->step;
            OptTrace = pCtx->trace;
            return do_proc(pp, Cp);
        }
        pp = pp->next;
    }
    return error("unknown procedure");
}

int cmd_return(void)
{
    if (!expr(&Rc))
        return error("invalid expression");
    pCtx->line = NULL;  // @@@ ?
    return TRUE;
}

int cmd_if(void)
{
    VALUE   v;
    char    buf[20];

    if (!expr(&v))
        return error("invalid expression");
    if (v.typ != T_NUM)
        return error("invalid type");
    if (v.num == 0)
        return TRUE;
    skipgap();
    if (!getname(buf, 19, TRUE))
        return error("keyword expected");
    if (!strcmp(buf, "GOTO"))
        return cmd_goto();
    if (!strcmp(buf, "DO"))
        return cmd_do();
    if (!strcmp(buf, "RETURN"))
        return cmd_return();
    if (!strcmp(buf, "EXIT"))
        return cmd_exit();
    return error("unknown keyword");
}

int do_while(char *cmd)
{
    VALUE   v;
    char    buf[20];

    for (;;)
    {
        Cp = cmd;
        if (!num_expr(&v))
            return error("invalid expression");
        if (v.num == 0)
            return TRUE;
        skipgap();
        if (!getname(buf, 19, TRUE))
            return error("keyword expected");
        if (!strcmp(buf, "DO"))
            if (!cmd_do())
                return FALSE;
    }
}

int cmd_while(void)
{
    int     rc;
    char    *sp;

    skipgap();
    sp = push_str(Cp);
    if (sp == NULL)
        return error("out of memory");
    rc = do_while(sp);
    free(sp);
    return rc;
}

int cmd_repeat(void)
{
    int     rc, savcnt;
    char    *sp;
    VALUE   v;

    savcnt = Cnt.num;
    skipgap();
    if (!num_expr(&v))
        return error("invalid expression");
    skipgap();
    sp = push_str(Cp);
    if (sp == NULL)
        return error("out of memory");
    for (Cnt.num = 1; Cnt.num <= v.num; Cnt.num++)
    {
        Cp = sp;
        rc = cmd_do();
        if (rc == 0)
            break;
    }
    free(sp);
    Cnt.num = savcnt;
    return rc;
}

int cmd_exitif(void)
{
    VALUE   v;

    if (!expr(&v))
        return error("invalid expression");
    if (v.typ != T_NUM)
        return error("invalid type");
    if (v.num == 0)
        return TRUE;
    skipgap();
    if (*Cp != ':')
        return error("colon expected");
    Cp++;
    fprintf(stderr, "EXIT: %s\n", Cp);
    Exitflg = TRUE;
    return TRUE;
}

int cmd_exit(void)
{
    skipgap();
    if (*Cp != ':')
        return error("colon expected");
    Cp++;
    fprintf(stderr, "EXIT: %s\n", Cp);
    Exitflg = TRUE;
    return TRUE;
}

int cmd_requires(void)
{
    int     ver, rel, rev;

    skipgap();
    if (!getdec(&ver)
    ||  !needchar('.')
    ||  !getdec(&rel)
    ||  !needchar('-')
    ||  !getdec(&rev))
        return error("invalid version string");
    if ((ver > LPU_VER)
    ||  (ver == LPU_VER && rel > LPU_REL)
    ||  (ver == LPU_VER && rel == LPU_REL && rev > LPU_REV))
        return error("newer lpu version required!");
    return TRUE;
}

int cmd_eps_insert(void)
{
    int     rc, sizflg = 0, sclflg = 0;
    int     vskip = 0, hskip = 0;
    int     hcenter = 0, vcenter = 0;
    char    fn[65], opt[11];
    tSIZE   size;               /* desired size of the image in DOTs       */
    tSIZE   esiz;               /* size of the original eps image in DOTs  */
    tPOS    pos;                /* desired position of the image in DOTs   */
    tSCALE  scale;
    VALUE   v;

    if (DevEPSInUse())
        return error("another eps file is open");

    pos.xpos = W->xpos;
    pos.ypos = W->ypos;
    skipgap();
    if (!getfname(fn, 64))
        return error("filename expected");
    skipgap();
    if (*Cp == '@')
    {
        Cp += 1;
        if (!get_position(&pos.xpos, &pos.ypos))
            return 0;
    }
    skipgap();
    while (*Cp == '.')
    {
        Cp += 1;
        if (!getname(opt, 10, FALSE))
            return error("option name expected");
        if (!strcmp(opt, "size")) {
            if (!get_size(&size))
                return 0;
            sizflg = 1;
        } else if (!strcmp(opt, "scale")) {
            if (!needlpar())
                return error("'(' expected");
            if (!flt_expr(&v))
                return error("x-scale expected");
            scale.xscale = v.flt;
            skipgap();
            if (!needcomma())
                return error("comma expected");
            skipgap();
            if (!flt_expr(&v))
                return error("y-scale expected");
            scale.yscale = v.flt;
            if (!needrpar())
                return error("')' expected");
            sclflg = 1;
        } else if (!strcmp(opt, "vskip")) {
            vskip = 1;
        } else if (!strcmp(opt, "hskip")) {
            hskip = 1;
        } else if (!strcmp(opt, "hcenter")) {
            hcenter = 1;
        } else if (!strcmp(opt, "vcenter")) {
            vcenter = 1;
        } else if (!strcmp(opt, "center")) {
            hcenter = 1;
            vcenter = 1;
        } else if (!strcmp(opt, "full")) {
            size.xsiz = W->xsiz;
            size.ysiz = W->ysiz;
            sizflg = 1;
        } else {
            return error("unknown option '%s'", opt);
        }
        skipgap();
    }

    if (sizflg && sclflg)
        return error("size OR scale!");

    rc = DevEPSOpen(fn, &esiz);
    if (sizflg == 0)
    {
        size.xsiz = esiz.xsiz;
        size.ysiz = esiz.ysiz;
    }
    else
    {
        rc = DevEPSResize(&size);
    }

    if (sclflg)
        rc = DevEPSRescale(&scale);

    if (hcenter)
        pos.xpos -= (size.xsiz / 2);
    if (vcenter)
        pos.ypos -= (size.ysiz / 2);
    rc = DevEPSAt(&pos);
    rc = DevEPSPrint();
    rc = DevEPSClose();

    if (vskip > 0)
    {
        W->ypos += size.ysiz;
    }
    if (hskip > 0)
    {
        W->xpos += size.xsiz;
    }

    return 1;
}

int cmd_eps_open(void)
{
    int     rc;
    char    fn[65];
    tSIZE   size;

    if (DevEPSInUse())
        return error("another eps file is open");

    skipgap();
    if (!getfname(fn, 64))
        return error("filename expected");

    rc = DevEPSOpen(fn, &size);
    if (rc == 0)
        return error("unable to read eps file");

    return 1;
}

int cmd_eps_close(void)
{
    int     rc;

    if (!DevEPSInUse())
        return error("no eps file open");

    rc = DevEPSClose();

    return rc;
}

int cmd_eps_resize(void)
{
    int     rc;
    tSIZE   size;

    if (!DevEPSInUse())
        return error("no eps file open");

    skipgap();
    if (!getunit(&size.xsiz))
        return error("x-size expected");
    skipgap();
    if (!needcomma())
        return error("comma expected");
    skipgap();
    if (!getunit(&size.ysiz))
        return error("y-size expected");

    rc = DevEPSResize(&size);

    return rc;
}

int cmd_eps_rescale(void)
{
    int     rc;
    tSCALE  scale;
    VALUE   v;

    if (!DevEPSInUse())
        return error("no eps file open");

    skipgap();
    //if (!getflt(&scale.xscale))
    if (!flt_expr(&v))
        return error("x-scale expected");
    scale.xscale = v.flt;
    skipgap();
    if (!needcomma())
        return error("comma expected");
    skipgap();
    //if (!getflt(&scale.yscale))
    if (!flt_expr(&v))
        return error("y-scale expected");
    scale.yscale = v.flt;

    rc = DevEPSRescale(&scale);

    return rc;
}

int cmd_eps_show(void)
{
    int     rc, vskip = 0, hskip = 0;
    int     hcenter = 0, vcenter = 0;
    char    opt[11];
    tPOS    pos;
    tSIZE   size;

    if (!DevEPSInUse())
        return error("no eps file open");

    pos.xpos = W->xpos;
    pos.ypos = W->ypos;
    skipgap();
    if (*Cp == '@')
    {
        Cp += 1;
        if (!get_position(&pos.xpos, &pos.ypos))
            return 0;
    }
    skipgap();
    while (*Cp == '.')
    {
        Cp += 1;
        if (!getname(opt, 10, FALSE))
            return error("option name expected");
        if (!strcmp(opt, "vskip")) {
            vskip = 1;
        } else if (!strcmp(opt, "hskip")) {
            hskip = 1;
        } else if (!strcmp(opt, "hcenter")) {
            hcenter = 1;
        } else if (!strcmp(opt, "vcenter")) {
            vcenter = 1;
        } else if (!strcmp(opt, "center")) {
            hcenter = 1;
            vcenter = 1;
        } else {
            return error("unknown option '%s'", opt);
        }
        skipgap();
    }

    DevEPSGetSize(&size);
    if (hcenter)
        pos.xpos -= (size.xsiz / 2);
    if (vcenter)
        pos.ypos -= (size.ysiz / 2);
    rc = DevEPSAt(&pos);
    rc = DevEPSPrint();

    if (vskip > 0)
    {
        W->ypos += size.ysiz;
    }
    if (hskip > 0)
    {
        W->xpos += size.xsiz;
    }

    return rc;
}

int cmd_eps(void)
{
    char    buf[21];

    skipgap();
    if (!getname(buf, 20, FALSE))
        return error("keyword expected");
    if (!strcmp(buf, "insert"))
        return cmd_eps_insert();
    if (!strcmp(buf, "open"))
        return cmd_eps_open();
    if (!strcmp(buf, "close"))
        return cmd_eps_close();
    if (!strcmp(buf, "resize"))
        return cmd_eps_resize();
    if (!strcmp(buf, "rescale"))
        return cmd_eps_rescale();
    if (!strcmp(buf, "show"))
        return cmd_eps_show();
    return error("invalid eps command");
}

int cmd_cr(void)
{
    W->xpos = W->lm;
    return 1;
}

int cmd_home(void)
{
    W->xpos = W->lm;
    W->ypos = W->tm;
    return 1;
}

int cmd_vskip(void)
{
    int     val;

    skipgap();
    if (!getunit(&val))
        return error("vskip value required");
    W->ypos += val;
    return 1;
}

int cmd_hskip(void)
{
    int     val;

    skipgap();
    if (!getunit(&val))
        return error("hskip value required");
    W->xpos += val;
    return 1;
}

int do_line(char *str)
{
    int     rc = TRUE;

    Cp = str;
    convert_line(Tmplin, 255);
    Cp = Tmplin;
    skipgap();
    if (pCtx->trace || pCtx->step)
        do_trace();
    if (pCtx->step)
        do_step();
    switch (*Cp)
    {
        case 0:
        case ';':
        case ':':
            break;
        case '.':
            Cp++;
            rc = do_cmd();
            break;
        case '"':
            rc = do_txt();
            do_nl(1);
            break;
        default:
            return error("invalid line format");
    }
    return 1;
}

static int get_params(tPROC *proc, char *cmdlin)
{
    VALUE   *p;

    proc->parcnt = 0;
    proc->params = NULL;
    Cp = cmdlin;
    skipgap();
    if (*Cp == 0 || *Cp == ';')
        return 1;
    do
    {
        if (proc->params == NULL)
            proc->params = (VALUE *) calloc(sizeof(VALUE), 1);
        else
            proc->params = (VALUE *) realloc(proc->params, sizeof(VALUE) * (proc->parcnt + 1));
        if (proc->params == NULL)
            return error("out of memory");
        p = proc->params + proc->parcnt;
        if (!expr(p))
            return error("expression error");
        proc->parcnt++;
        skipgap();
    } while (needcomma());
    return 1;
}

int do_proc(tPROC *proc, char *cmdlin)
{
    int     rc = TRUE;

    Rc.num = -1;
    if (Ctxcnt >= MAXNEST)
        return error("include nesting too deep");
    Rc.num = 0;
    if (!get_params(proc, cmdlin))
        return FALSE;
    Ctxcnt++;
    pCtx = &Ctxtab[Ctxcnt];
    pCtx->proc = proc;
    pCtx->line = proc->first;
    pCtx->vars.first = NULL;
    pCtx->vars.last  = NULL;
    pCtx->step  = OptStep;
    pCtx->trace = OptTrace;
    if (pCtx->trace || pCtx->step)
        fprintf(stderr, "enter proc '%s', level=%d\n",
            proc->procname, Ctxcnt);
    while (rc && pCtx->line != NULL && !Exitflg)
    {
        rc = do_line(pCtx->line->line);
        pCtx->line = pCtx->line->next;
    }
    if (pCtx->trace || pCtx->step)
        fprintf(stderr, "leave proc '%s'\n", proc->procname);
    symDelAll(&(pCtx->vars));
    Ctxcnt--;
    if (Ctxcnt >= 0)
        pCtx--;
    else
        pCtx = NULL;
    return TRUE;
}

static int   srcmode = -1;
static tPROC *proc = &Main;
static tLINE *line = NULL;

static int add_proc(char *buf)
{
    tPROC   *pp;

    while (*buf == ' ')
        buf++;
    if (*buf == 0)
        return error("procname expected");
    pp = calloc(1, sizeof(tPROC));
    if (pp == NULL)
        return error("out of memory");
    pp->next     = NULL;
    pp->first    = NULL;
    pp->procname = push_str(buf);
    if (pp->procname == NULL)
        return error("out of memory");
    proc->next = pp;
    proc = pp;
    line = NULL;
    return TRUE;
}

static int add_line(char *buf)
{
    tLINE   *lp;

    lp = calloc(1, sizeof(tLINE));
    if (lp == NULL)
        return error("out of memory");
    lp->next = NULL;
    lp->line = push_str(buf);
    if (lp->line == NULL)
        return error("out of memory");
    if (line == NULL)
        proc->first = lp;
    else
        line->next  = lp;
    line = lp;
    return TRUE;
}

static void dump_proc(tPROC *proc)
{
    tLINE   *lp;

    fprintf(stderr, "Procname=<%s>\n", proc->procname);
    lp = proc->first;
    while (lp != NULL)
    {
        fprintf(stderr, "Line='%s'\n", lp->line);
        lp = lp->next;
    }
    fprintf(stderr, "End\n");
}

int pp_include(void)
{
    char    fn[80];

    skipgap();
    if (*Cp == 0)
        error("filename expected");
    strcpy(fn, Cp);
    if (strchr(fn, '.') == NULL)
        strcat(fn, ".lpi");
    if (!do_include(fn))
        return FALSE;
    return TRUE;
}

int preproc(void)
{
    char    cmd[20];

    Cp++;
    if (!getname(cmd, 19, FALSE))
        return error("preproc-command expected");
    if (!strcmp(cmd, "include"))        return pp_include();
    return error("unknown preproc-command");
}

static int valid_comment(char *p)
{
    while (*p == ' ' | *p == '\t')
        p += 1;
    if (*p == ';')
        return 1;
    return 0;
}

int do_include(char *fn)
{
    FILE    *fp;
    int     rc = TRUE;

    if (fn == NULL)
        fp = stdin;
    else
    {
        fp = fopen(fn, "r");
        if (fp == NULL)
            return error("unable to open includefile '%s'", fn);
    }
    while (rc && nextline(fp, Wrklin, 255) && !Exitflg)
    {
        switch (Wrklin[0])
        {
            case '#':
                if (!isalpha(Wrklin[1]))
                    break;
                rc = preproc();
                break;
            case ':':
                if (Wrklin[1] == ':')
                {
                    rc = add_proc(&Wrklin[2]);
                    break;
                }
                rc = add_line(Wrklin);
                break;
            case '.':
            case '"':
                rc = add_line(Wrklin);
                break;
            case ';':
            case 0:
                break;
            default:
                if (valid_comment(Wrklin))
                    break;
                error("invalid line '%s'", Wrklin);
                rc = FALSE;
                break;
        }
    }
    if (fn != NULL)
        fclose(fp);

    return rc;
}


int do_cmd(void)
{
    char    buf[20];

    if (!getname(buf, 19, TRUE))
        return error("command expected");

    if (!strcmp(buf, "REQUIRES"))       return cmd_requires();

    if (!strcmp(buf, "DEFINK"))         return cmd_defink();
    if (!strcmp(buf, "DEFPEN"))         return cmd_defpen();
    if (!strcmp(buf, "DEFSTYLE"))       return cmd_defstyle();

    if (!strcmp(buf, "FORM"))           return cmd_form();
    if (!strcmp(buf, "LABEL"))          return cmd_label();
    if (!strcmp(buf, "EJECT"))          return cmd_eject(1);
    if (!strcmp(buf, "NOEJECT"))        return cmd_eject(0);
    if (!strcmp(buf, "FONT"))           return cmd_font();
    if (!strcmp(buf, "STYLE"))          return cmd_style();
    if (!strcmp(buf, "PEN"))            return cmd_pen();
    if (!strcmp(buf, "INK"))            return cmd_ink();
    if (!strcmp(buf, "VAR"))            return cmd_var();
    if (!strcmp(buf, "LET"))            return cmd_let();
    if (!strcmp(buf, "IF"))             return cmd_if();
    if (!strcmp(buf, "WHILE"))          return cmd_while();
    if (!strcmp(buf, "REPEAT"))         return cmd_repeat();
    if (!strcmp(buf, "GOTO"))           return cmd_goto();
    if (!strcmp(buf, "DO"))             return cmd_do();
    if (!strcmp(buf, "RETURN"))         return cmd_return();
    if (!strcmp(buf, "EXIT"))           return cmd_exit();
    if (!strcmp(buf, "EXITIF"))         return cmd_exitif();
    if (!strcmp(buf, "DEBUG"))          return cmd_debug();
    if (!strcmp(buf, "STOP"))           return cmd_stop();
    if (!strcmp(buf, "CONT"))           return cmd_cont();
    if (!strcmp(buf, "TRON"))           return cmd_tron();
    if (!strcmp(buf, "TROFF"))          return cmd_troff();
    if (!strcmp(buf, "WATCH"))          return cmd_watch();
    if (!strcmp(buf, "INPUT"))          return cmd_input();
    if (!strcmp(buf, "DB"))             return cmd_db();
    if (!strcmp(buf, "USE"))            return cmd_use();

    if (pLbl == NULL)
        return error("no label defined");

    if (!strcmp(buf, "WINDOW"))         return cmd_window();
    if (!strcmp(buf, "BORDER"))         return cmd_border();
    if (!strcmp(buf, "BASE"))           return cmd_base();
    if (!strcmp(buf, "SETXY"))          return cmd_setxy();
    if (!strcmp(buf, "HLINE"))          return cmd_hline();
    if (!strcmp(buf, "NL"))             return cmd_nl();
    if (!strcmp(buf, "PRINT"))          return cmd_print(0);
    if (!strcmp(buf, "UPRINT"))         return cmd_print(1);
    if (!strcmp(buf, "TEXT"))           return cmd_text(Txtadj);
    if (!strcmp(buf, "LTEXT"))          return cmd_text(LEFT);
    if (!strcmp(buf, "RTEXT"))          return cmd_text(RIGHT);
    if (!strcmp(buf, "CTEXT"))          return cmd_text(CENTER);
    if (!strcmp(buf, "LEFT"))           return cmd_left();
    if (!strcmp(buf, "RIGHT"))          return cmd_right();
    if (!strcmp(buf, "CENTER"))         return cmd_center();
    if (!strcmp(buf, "FRAME"))          return cmd_frame();
    if (!strcmp(buf, "GBOX"))           return cmd_gbox();
    if (!strcmp(buf, "GLINE"))          return cmd_gline();
    if (!strcmp(buf, "GCIRCLE"))        return cmd_gcircle();
    if (!strcmp(buf, "CR"))             return cmd_cr();
    if (!strcmp(buf, "HOME"))           return cmd_home();
    if (!strcmp(buf, "HSKIP"))          return cmd_hskip();
    if (!strcmp(buf, "VSKIP"))          return cmd_vskip();

    if (!strcmp(buf, "EPS"))            return cmd_eps();

    return error("unknown command '%s'",buf);
}


/***** Konfiguration der Formulardaten *****/

static int cfgflg;
static char cfgbuf[L_NAME+1];

int tab_cfg_form(void)
{
    if (Frmcnt >= MAXFRM)
        return error("too many forms");
    cfgflg = 1;
    Form = &Forms[Frmcnt];
    skipgap();
    if (!getname(Form->id, L_ID, FALSE))
        return error("formid expected");
    skipgap();
    if (!getname(Form->page, L_ID, FALSE))
        return error("pageid expected");
    Form->lblcnt = 0;
    Form->lbl1st = -1;
    Frmcnt++;
    return TRUE;
}

int tab_cfg_label(void)
{
    if (cfgflg < 1)
        return error("unexpected label-directive");
    if (Lblcnt >= MAXLBL)
        return error("too many labels");
    pLbl   = &Labels[Lblcnt];
    cfgflg = 2;
    skipgap();
    if (!getname(pLbl->id, L_ID, FALSE))
        return error("label-id expected");
    skipgap();
    if (!getunit(&pLbl->xsize))
        return error("xsize expected");
    skipgap();
    if (!getunit(&pLbl->ysize))
        return error("ysize expected");
    pLbl->xbase = 0;
    pLbl->ybase = 0;
    pLbl->xcnt  = 1;
    pLbl->ycnt  = 1;
    if (Form->lbl1st < 0)
        Form->lbl1st = Lblcnt;
    Form->lblcnt++;
    Lblcnt++;
    return TRUE;
}

int tab_cfg_base(void)
{
    if (cfgflg != 2)
        return error("unexpected base-directive");
    skipgap();
    if (!getunit(&pLbl->xbase))
        return error("xbase expected");
    skipgap();
    if (!getunit(&pLbl->ybase))
        return error("ybase expected");
    return TRUE;
}

int tab_cfg_xrep(void)
{
    if (cfgflg != 2)
        return error("unexpected xrep-directive");
    skipgap();
    if (!getdec(&pLbl->xcnt))
        return error("xcnt expected");
    skipgap();
    if (!getunit(&pLbl->xskip))
        return error("xskip expected");
    return TRUE;
}

int tab_cfg_yrep(void)
{
    if (cfgflg != 2)
        return error("unexpected yrep-directive");
    skipgap();
    if (!getdec(&pLbl->ycnt))
        return error("ycnt expected");
    skipgap();
    if (!getunit(&pLbl->yskip))
        return error("yskip expected");
    return TRUE;
}

int tab_cfg_item(void)
{
    if (!getname(cfgbuf, L_NAME, FALSE))
        return error("keyword expected");
    if (!strcmp(cfgbuf, "form"))
        return tab_cfg_form();
    if (!strcmp(cfgbuf, "label"))
        return tab_cfg_label();
    if (!strcmp(cfgbuf, "base"))
        return tab_cfg_base();
    if (!strcmp(cfgbuf, "xrep"))
        return tab_cfg_xrep();
    if (!strcmp(cfgbuf, "yrep"))
        return tab_cfg_yrep();
    return error("unknown keyword '%s'", cfgbuf);
}

int tab_cfg(void)
{
    char    fn[65];
    FILE    *fp;
    int     i;

    verbose(1, "table-configuration...\n");
    sprintf(fn, "%s", Tabfile);
    fp = fopen(fn, "r");
    if (fp == NULL && fn[0] != '/')
    {
        sprintf(fn, "%s/.lpu/%s", Usrdir, Tabfile);
        fp = fopen(fn, "r");
        if (fp == NULL)
        {
            sprintf(fn, "%s/%s", Lpudir, Tabfile);
            fp = fopen(fn, "r");
            if (fp == NULL)
                return error("tabfile '%s' not found",Tabfile);
        }
    }
    verbose(1, "using tabfile '%s'...\n", fn);
    cfgflg = 0;
    while (nextline(fp, Tmplin, 255) && cfgflg >= 0)
    {
        skipgap();
        switch (*Cp)
        {
            case ';':
            case 0:
                break;
            default:
                tab_cfg_item();
                break;
        }
    }
    fclose(fp);
    if (cfgflg == 0)
        return error("no forms defined");
    for (i=0; i<Frmcnt; i++)
    {
        if (Forms[i].lblcnt == 0)
            return error("no label for form '%s' defined", Forms[i].id);
    }
    return TRUE;
}


int par_cfg_item(void)
{
    char    buf[L_ID+1];

    if (!getname(buf, L_ID, FALSE))
        return error("keyword expected");
    return error("unknown item '%s'", buf);
}

int par_cfg_dotcmd(void)
{
    char    buf[L_ID+1];

    if (!getname(buf, L_ID, FALSE))
        return error("keyword expected");
    if (!strcmp(buf, "defink"))
        return cmd_defink();
    if (!strcmp(buf, "defpen"))
        return cmd_defpen();
    if (!strcmp(buf, "defstyle"))
        return cmd_defstyle();
    return error("unknown keyword '%s'", buf);
}

int par_cfg(char *parfil, int needed)
{
    char    fn[65];
    FILE    *fp;
    int     i;

    verbose(1, "paramfile-configuration...\n");
    sprintf(fn, "%s", parfil);
    fp = fopen(fn, "r");
    if (fp == NULL && fn[0] != '/')
    {
        sprintf(fn, "%s/.lpu/%s", Usrdir, parfil);
        fp = fopen(fn, "r");
        if (fp == NULL)
        {
            sprintf(fn, "%s/%s", Lpudir, parfil);
            fp = fopen(fn, "r");
            if (fp == NULL)
                if (needed)
                    return error("parfile '%s' not found", parfil);
                else
                    return TRUE;
        }
    }
    verbose(1, "using parfile '%s'...\n", fn);
    while (nextline(fp, Tmplin, 255))
    {
        skipgap();
        switch (*Cp)
        {
            case ';':
            case 0:
                break;
            case '.':
                Cp++;
                par_cfg_dotcmd();
                break;
            default:
                par_cfg_item();
                break;
        }
    }
    fclose(fp);
    return TRUE;
}


void lpu_init(void)
{
    char    *p;

    strcpy(Devfile, "lpu.dev");
    strcpy(Tabfile, "lpu.tab");
    strcpy(Parfile, "lpu.par");
    strcpy(Outfile, "");
    strcpy(Usrdir, "~");
    strcpy(Lpudir, "~");
    strcpy(Afmdir, "~");
    p = getenv("HOME");
    if (p != NULL)
        strcpy(Usrdir, p);
    p = getenv("LPUDIR");
    if (p != NULL)
        strcpy(Lpudir, p);
    OptFrame   = 0;
    OptVerbose = 0;
    OptQuiet   = 0;
    OptTrace   = 0;
    OptStep    = 0;
    OptBatch   = 0;
    OptStdin   = 1;

    Frmcnt = 0;
    Lblcnt = 0;
    Ofp    = stdout;

    Par.x_unit = 0;
    Par.y_unit = 0;
    Par.t_unit = 0;
    Par.r_unit = 0;

    Main.next     = NULL;
    Main.first    = NULL;
    Main.procname = "main";

    Gsyms.first   = NULL;
    Gsyms.last    = NULL;

    Cnt.typ = T_NUM;
    Cnt.num = 0;
}

int do_arg(char *arg)
{
    char    opt;

    opt = *arg++;
    switch (opt)
    {
        case 'b':
            OptBatch = 1;
            break;
        case 'd':
            strcpy(Devfile, arg);
            strcat(Devfile, ".dev");
            break;
        case 's':
            {
                char    *p;
                SYMBOL  *sym;

                p = strchr(arg, '=');
                if (p != NULL)
                    *p++ = 0;
                sym = symAdd(&Gsyms, arg);
                if (sym == NULL)
                {
                    fprintf(stderr, "unable to define symbol '%s'\n",
                        arg);
                    break;
                }
                if (p != NULL)
                {
                    sym->val.typ = T_STR;
                    strcpy(sym->val.str, p);
                }
            }
            break;
        case 't':
            strcpy(Tabfile, arg);
            strcat(Tabfile, ".tab");
            break;
        case 'p':
            strcpy(Parfile, arg);
            strcat(Parfile, ".par");
            break;
        case 'o':
            strcpy(Outfile, arg);
            break;
        case 'f':
            OptFrame = 1;
            break;
        case 'v':
            OptVerbose = 1;
            if (isdigit(*arg))
                OptVerbose = *arg - '0';
            break;
        case 'j':
            Skipcnt = atoi(arg);
            break;
        case '-':
            OptStdin = 1;
            break;
        case 'q':                       /* already processed               */
            break;
        default:
            fprintf(stderr, "invalid option '%c%s'\n",opt, arg);
            return 0;
    }
    return 1;
}

int do_argfile(char *fn)
{
    int     rc;
    char    buf[256];
    FILE    *fp;

    fp = fopen(fn, "r");
    if (fp == NULL)
    {
        fprintf(stderr, "unable to open optionfile '%s'\n", fn);
        return 0;
    }
    while (!feof(fp))
    {
        fgets(buf, 255, fp);
        strip_lf(buf);
        rc = do_arg(buf);
    }
    fclose(fp);
    return 1;
}

void usage(char *prog)
{
    fprintf(stderr, "Usage: %s { option } srcfile ...\n", prog);
    fprintf(stderr, "  option: @optfile            use an option file\n");
    fprintf(stderr, "          --                  dummy option\n");
    fprintf(stderr, "          -b                  batch mode\n");
    fprintf(stderr, "          -d<devfile>         use different dev file\n");
    fprintf(stderr, "          -f                  draw frame around labels\n");
    fprintf(stderr, "          -h                  display this help screen\n");
    fprintf(stderr, "          -j<count>           skip <count> labels\n");
    fprintf(stderr, "          -o<outfile>         output to given file\n");
    fprintf(stderr, "          -p<parfile>         use different param file\n");
    fprintf(stderr, "          -q                  be quiet\n");
    fprintf(stderr, "          -s<symbol>=<value>  predefine a symbol\n");
    fprintf(stderr, "          -t<tabfile>         use different table file\n");
    fprintf(stderr, "          -v[level]           verbosity level [0-9]\n");
}

void print_version(void)
{
    sprintf(Version, "%d.%d-%d%s", LPU_VER, LPU_REL, LPU_REV, LPU_STP);
    fprintf(stderr, "LPU -- Label Printing Utility v%s, build date %s\n\n",
        Version, GENDATE);
}

int main(int argc, char *argv[])
{
    int     ac, rc;

    lpu_init();

    if (argc == 1)
    {
        print_version();
        usage(argv[0]);
        exit(0);
    }
    for (ac=1; ac<argc; ac++)
    {
        if (!strcmp(argv[ac], "-h"))
        {
            print_version();
            usage(argv[0]);
            exit(0);
        }
        if (!strcmp(argv[ac], "-q"))
        {
            OptQuiet = 1;
        }
    }

    if (!OptQuiet)
        print_version();
    verbose(1, "Usrdir='%s', Lpudir='%s'\n", Usrdir, Lpudir);

    signal(SIGINT, sighandler);

    for (ac=1; ac<argc; ac++)
    {
        if (argv[ac][0] == '-')
        {
            rc = do_arg(&argv[ac][1]);
        }
        else
        {
            if (argv[ac][0] == '@')
                rc = do_argfile(&argv[ac][1]);
            else
            {
                OptStdin = 0;
                rc = 1;
            }
        }
    }

    tdb_init();

    DevReset();

    rc = DevConfig(Devfile);
    if (rc < 0) {
        fprintf(stderr, "DevConfig(%s) failed with rc %d\n", Devfile, rc);
        return 1;
    }

    if (Outfile[0] > ' ')
        if ((Ofp = fopen(Outfile, "w")) == NULL) {
            fprintf(stderr, "unable to open outputfile '%s'\n",Outfile);
            return 1;
        }

    DevBeginJob(Ofp);

    if (!tab_cfg())
        return 1;
    verbose(1, "nach tab_cfg()\n");

    Form = NULL;   /* no Form selected */
    pLbl = NULL;   /* no Label selected */
    DevSelFont(NULL);

    if (!par_cfg(Parfile, FALSE))
        return 1;
    verbose(1, "nach par_cfg(%s)\n", Parfile);

    Ctxcnt = -1;
    pCtx   = NULL;
    if (OptStdin)
    {
        fprintf(stderr, "LPU is running in STDIN mode.\n");
        do_include(NULL);
    }
    else
    {
        for (ac=1; ac<argc; ac++)
        {
            if (argv[ac][0] != '-' && argv[ac][0] != '@')
            {
                strcpy(Cmdfile, argv[ac]);
                if (strchr(Cmdfile, '.') == NULL)
                    strcat(Cmdfile, ".lpu");
                do_include(Cmdfile);
            }
        }
    }

    do_proc(&Main, "");

    if (pLbl != NULL)
        DevEndPage(1);
    else
        DevEndPage(0);

    DevEndJob();
    DevExit();
    tdb_exit();
    if (Ofp != NULL && Outfile[0] > ' ')
        fclose(Ofp);
    return 0;
}
