/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 *      lpudefs.h --    Definitionen fuer LPU
 *
 *          jrs     18.07.96
 */

#ifndef lpudefs_h
#define lpudefs_h

#include "expr.h"

#define LEFT	0
#define RIGHT	1
#define CENTER	2

#define MAXNEST     5
#define MAXLBL      100
#define MAXFRM      50
#define MAXWIN      10

typedef struct line
{
    struct line     *next;
    char            *line;
} tLINE;

typedef struct proc
{
    struct proc     *next;
    char            *procname;
    tLINE           *first;
    int             parcnt;
    VALUE           *params;
} tPROC;

typedef struct
{
    int     lnr;                        /* Zeilennummer im Modul            */
    tPROC   *proc;
    tLINE   *line;
    tVARS   vars;
    int     trace;
    int     step;
} tCTX;

typedef struct              /* Referenz auf einen Labeltyp fuer eine Form   */
{
    char    id[L_ID+1];                 /* Label-ID                         */
    int     xsize;                      /* width                            */
    int     ysize;                      /* height                           */
    int     xbase;                      /* base-offset X                    */
    int     ybase;                      /* base-offset Y                    */
    int     xskip;                      /* distance to next label X         */
    int     yskip;                      /* distance to next label Y         */
    int     xcnt;                       /* Anzahl Label nebeneinander       */
    int     ycnt;                       /* Anzahl Label untereinander       */
    int     xlab;                       /* aktueller Label horizontal       */
    int     ylab;                       /* aktueller Label vertikal         */
    int     ndx;                        /* aktuelle Labelnummer             */
} tLABEL;

typedef struct              /* Aufbau einer Seite, alle Werte in 1/720"     */
{
    char    id[L_ID+1];                 /* Etiketten-ID                     */
    char    page[L_ID+1];               /* Seiten-ID                        */
    int     lbl1st;                     /* Zeiger in Label-Reftab           */
    int     lblcnt;                     /* Anzahl der Label-Referenzen      */
} tFORM;

typedef struct              /* Arbeitsparameter fuer LPU                    */
{
    int     xoff;               /* aktueller BASE-Wert, relativ zum Label   */
    int     yoff;
    int     xsiz;               /* aktuelle Groesse des Labels              */
    int     ysiz;
    int     xpos;               /* aktuelle Schreibposition, relativ        */
    int     ypos;
    int     xref;               /* Referenzwert                             */
    int     yref;
    int     tm;                 /* die Raender                              */
    int     bm;
    int     lm;
    int     rm;
} tWIN;

typedef struct              /* allgemeine Parameter                         */
{
    int     t_unit;             /* aktuelle Zeichenbreite                   */
    int     r_unit;             /* aktuelle Zeichenhoehe                    */
    int     x_unit;             /* aktuelle X-Einheit                       */
    int     y_unit;             /* aktuelle Y-Einheit                       */
} tPAR;

typedef struct      /* Farbdefinition CMYK                              */
{
    int     cyan;
    int     magenta;
    int     yellow;
    int     black;
} tCMYK;

typedef struct      /* Styledefinition                                  */
{
    char    family[L_ID+1];
    int     points;
    int     attr;
} tSTYLE;

typedef struct item         /* ITEM-Definitionen                            */
{
    struct item *next;
    char        id[L_ID+1];
    int         pen;
    tCMYK       ink;
    tSTYLE      style;
} ITEM;

typedef struct
{
    int     xpos;
    int     ypos;
} tPOS;

typedef struct
{
    int     xsiz;
    int     ysiz;
} tSIZE;

typedef struct
{
    double  xscale;
    double  yscale;
} tSCALE;

#endif	/* lpudefs_h */
