/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 * tdbfuncs.c -- Textorientierte Datenbankfunktionen
 *
 *  JRS 30.07.97
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "utils.h"
#include "tdbfuncs.h"

static	tTDB	Dbtab[MAXTDB+1];
static	tTDB	*Dbu;

int tdb_init(void)
{
    int     i;

    Dbu = &Dbtab[1];
    for (i=0; i<=MAXTDB; i++)
    {
        Dbtab[i].fp = NULL;
        // @@@
    }
    return 1;
}

int tdb_exit(void)
{
    int     i;

    for (i=0; i<=MAXTDB; i++)
    {
        if (Dbtab[i].fp != NULL)
        {
            fclose(Dbtab[i].fp);
            Dbtab[i].fp = NULL;
        }
    }
    return 1;
}

int tdb_use(int n)
{
    if ((n < 1) || (n > MAXTDB))
        return 0;
    Dbu = &Dbtab[n];
    return 1;
}

int tdb_open(int n, char *dbnam, int typ, int flags, char sep)
{
    int     ndx;

    if (typ != 0)
        return 0;
    if (flags != 0)
        return 0;
    if (n > 0)
    {
        if (n > MAXTDB)
            return 0;
        Dbu = &Dbtab[n];
    }
    if (Dbu->fp != NULL)
        return 0;
    strcpy(Dbu->name, dbnam);
    if (strchr(dbnam, '.') == NULL)
        strcat(Dbu->name, ".tdb");
    Dbu->fldsep = sep;
    Dbu->fp = fopen(Dbu->name, "r");
//fprintf(stderr, "tdb_open(%s) %s\n",
//  Dbu->name, Dbu->fp == NULL ? "failed" : "ok");
    if (Dbu->fp == NULL)
        return 0;
    Dbu->recno = 0;
    Dbu->eodflg = 1;
    return 1;
}

int tdb_close(void)
{
    if (Dbu->fp == NULL)
        return 0;
    fclose(Dbu->fp);
    Dbu->fp = NULL;
	Dbu->recno = 0;
    Dbu->eodflg = 1;
    return 1;
}

int tdb_top(void)
{
    if (Dbu->fp == NULL)
        return 0;
    rewind(Dbu->fp);
    Dbu->recno = 0;
    return tdb_nxtrec();
}

int tdb_next(void)
{
    if (Dbu->fp == NULL)
        return 0;
    if (Dbu->eodflg)
        return 1;
    return tdb_nxtrec();
}

int tdb_find(int f, char *id)
{
    int     found, fnr;

    if (Dbu->fp == NULL)
        return 0;
    if ((f < 1) || (f > MAXFLD))
        return 0;
    f--;
    found = 0;
    while (!Dbu->eodflg && !found)
    {
        if (!strcmp(id, Dbu->fldtab[f]))
            found = 1;
        else
            tdb_nxtrec();
    }
    return 1;
}

int tdb_append(int fil, char *data)
{
    return 1;
}

static int isfldsep(char c)
{
    if ((c == 0) || (c == Dbu->fldsep))
        return 1;
    return 0;
}

int tdb_nxtrec(void)
{
    int     ndx;
    char    quote;

    nextline(Dbu->fp, Dbu->line, 1023);
    for (ndx=0; ndx<MAXFLD; ndx++)
        Dbu->fldtab[ndx] = NULL;
    if (feof(Dbu->fp))
    {
        Dbu->eodflg = 1;
        Dbu->recno = 0;
        return 1;
    }
    Dbu->eodflg = 0;
    Dbu->recno++;
    Cp = Dbu->line;
    ndx = 0;
    while (*Cp && (ndx < MAXFLD))
    {
        switch (*Cp)
        {
            case 0x22:
            case 0x27:
                quote = *Cp++;
                Dbu->fldtab[ndx++] = Cp;
                while (*Cp)
                {
                    if ((*Cp == quote) && (isfldsep(*(Cp+1))))
                        break;
                    Cp++;
                }
                if (*Cp == 0)
                    return 0;
                *Cp++ = 0;
                break;
            default:
                Dbu->fldtab[ndx++] = Cp;
                while (*Cp && (*Cp != Dbu->fldsep))
                    Cp++;
                break;
        }
        if (*Cp == Dbu->fldsep)
            *Cp++ = 0;
    }
    Dbu->fldcnt = ndx;
    return 1;
}

static tTDB *use_tdb(int db)
{
    if (db == 0)
        return Dbu;
    if (db >= 1 && db <= MAXTDB)
        return &Dbtab[db];
    return NULL;
}

char *tdb_getfld(int db, int n)
{
    tTDB    *tdb;

    tdb = use_tdb(db);
    if (tdb == NULL)
        return NULL;
    if (n < 1 || n > tdb->fldcnt)
        return NULL;
    return tdb->fldtab[n-1]; // @@@ hier evtl. Feld 0 fuer gesamte Zeile!
}

int tdb_eof(int db)
{
    tTDB    *tdb;

    tdb = use_tdb(db);
    if (tdb == NULL)
        return -1;
    return tdb->eodflg;
}

int tdb_recno(int db)
{
    tTDB    *tdb;

    tdb = use_tdb(db);
    if (tdb == NULL)
        return -1;
    return tdb->recno;
}

int tdb_fldcnt(int db)
{
    tTDB    *tdb;

    tdb = use_tdb(db);
    if (tdb == NULL)
        return -1;
    return tdb->fldcnt;
}

