/*

    LPU -- Label Printing Utility
    Copyright (C) 1998  John R. Schrader

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

/*
 *  utils.c -- diverse Hilfsfunktionen
 *
 *      jrs     02.08.97
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "cfg.h"
#include "expr.h"
#include "utils.h"
#include "lpudefs.h"
#include "lpuvars.h"

char    *Cp;

int warning(char *fmt, ...)
{
    va_list arglst;
    char    buf[256];

    va_start(arglst, fmt);
    vsprintf(buf, fmt, arglst);
    va_end(arglst);
    fprintf(stderr,"warning: %s\n",buf);
    if (Ctxcnt >= 0 && pCtx != NULL)
    {
        fprintf(stderr, "  in proc '%s'\n", pCtx->proc->procname);
        fprintf(stderr, "    txt='%s'\n", pCtx->line->line);
        //fprintf(stderr, "    txt='%s'\n", Tmplin);
    }
    return TRUE;
}

int error(char *fmt, ...)
{
    va_list arglst;
    char    buf[256];

    va_start(arglst, fmt);
    vsprintf(buf, fmt, arglst);
    va_end(arglst);
    fprintf(stderr,"error: %s\n",buf);
    if (Ctxcnt >= 0 && pCtx != NULL)
    {
        fprintf(stderr, "  in proc '%s'\n", pCtx->proc->procname);
        fprintf(stderr, "    txt='%s'\n", pCtx->line->line);
        //fprintf(stderr, "    txt='%s'\n", Tmplin);
    }
    return FALSE;
}

void verbose(int level, char *fmt, ...)
{
    va_list arglst;

    if (OptVerbose >= level)
    {
        va_start(arglst, fmt);
        vfprintf(stderr, fmt, arglst);
        va_end(arglst);
    }
}

int nextline(FILE *fp, char buf[], int siz)
{
    char    *p;
    int     n, goon;

    Cp = buf;
    do
    {
        fgets(buf, siz, fp);
        p = buf;
        if (feof(fp)) { *p = 0; return 0; }
        n = 0; goon = 0;
        while (*p)
        {
            if (*p == '\n')
            {
                if (n > 0 && *(p-1) == '\\')
                {
                    n--; p--;
                    goon = 1;
                }
                *p = 0;
                break;
            }
            p++; n++;
        }
        buf = p;
        siz -= n;
    } while (goon);
    return TRUE;
}

void skipgap(void)
{
    while ((*Cp == ' ') || (*Cp == '\t'))
        Cp++;
}

int needchar(char c)
{
    if (*Cp != c)
        return 0;
    Cp++;
    return 1;
}

int needcomma(void)     { return needchar(','); }
int needdot(void)       { return needchar('.'); }
int needsemi(void)      { return needchar(';'); }
int needlpar(void)      { return needchar('('); }
int needrpar(void)      { return needchar(')'); }

int needstr(char *str)
{
    int     len = strlen(str);

    if (len < 1)
        return 0;
    if (memcmp(str, Cp, len))
        return 0;
    if (isalnum(Cp+len))
        return 0;
    Cp += len;
    return 1;
}

int needunit(double *unit)
{
    char    buf[5];

    if (!getname(buf, 4, 0))
        return 0;

    if (!strcmp(buf, "i"))      { *unit = 1.0 * DPI;         return 1; }
    if (!strcmp(buf, "c"))      { *unit = 1.0 * DPI / 2.54;  return 1; }
    if (!strcmp(buf, "m"))      { *unit = 1.0 * DPI / 25.4;  return 1; }
    if (!strcmp(buf, "p"))      { *unit = 1.0 * DPI / 72.0;  return 1; }
    if (!strcmp(buf, "r"))      { *unit = 1.0 * Par.r_unit;  return 1; }
    if (!strcmp(buf, "t"))      { *unit = 1.0 * Par.t_unit;  return 1; }
    if (!strcmp(buf, "x"))      { *unit = 1.0 * Par.x_unit;  return 1; }
    if (!strcmp(buf, "y"))      { *unit = 1.0 * Par.y_unit;  return 1; }

    return 0;
}

int getname(char buf[], int siz, int flg)
{
    int     i;
    char    c;

    if (!isalpha(*Cp))
        return 0;
    i = 0;
    while (isalnum(*Cp))
    {
        c = *Cp++;
        if (flg)
            c = toupper(c);
        if (i < siz)
            buf[i++] = c;
    }
    buf[i] = 0;
    return 1;
}

int getdec(int *x)
{
    int     s = 1;

    if (*Cp == '-') { s = -1; Cp++; }
    if (!isdigit(*Cp))
        return 0;
    *x = 0;
    while (isdigit(*Cp))
        *x = *x * 10 + *Cp++ - '0';
    *x = *x * s;
    return 1;
}

int getflt(double *x)
{
    int     s = 1;
    double  num, dec;

    num  = 0.;
    dec  = 1;
    if (*Cp == '-') { s = -1; Cp++; }
    if (!isdigit(*Cp))
        return 0;
    while (isdigit(*Cp))
        num = num * 10. + (*Cp++ - '0');
    if (*Cp == '.')
    {
        Cp++;
        while (isdigit(*Cp))
        {
            dec /= 10.;
            num = num + dec * (*Cp++ - '0');
        }
    }
    *x = s * num;
    return 1;
}

int getunit(int *val)
{
#if 1
    VALUE   v;

    if (!unit_expr(&v))
        return error("unit expected");
    *val = v.num;
#else
    double  num, dec;
    int     n, s = 1;
    VALUE   v;

    num  = 0.;
    dec  = 1;
    *val = 0;
    if (needchar('('))
    {
        if (!expr(&v))
            return 0;
        if (v.typ != T_NUM && v.typ != T_FLT)
            return 0;
        if (!needchar(')'))
            return 0;
        if (v.typ == T_NUM)
            v.flt = (double) 1.0 * v.num;
        num = v.flt;
    }
    else
    {
        if (*Cp == '-') { s = -1; Cp++; }
        if (!isdigit(*Cp))
            return 0;
        while (isdigit(*Cp))
            num = num * 10. + (*Cp++ - '0');
        if (*Cp == '.')
        {
            Cp++;
            while (isdigit(*Cp))
            {
                dec /= 10.;
                num = num + dec * (*Cp++ - '0');
            }
        }
    }
    switch (*Cp)
    {
        case 'i':   num = num * DPI;                    Cp++;   break;
        case 'c':   num = num * DPI / 2.54;             Cp++;   break;
        case 'm':   num = num * DPI / 25.4;             Cp++;   break;
        case 'p':   num = num * DPI / 72.;              Cp++;   break;
        case 'r':   num = num * Par.r_unit;             Cp++;   break;
        case 't':   num = num * Par.t_unit;             Cp++;   break;
        case 'x':   num = num * Par.x_unit;             Cp++;   break;
        case 'y':   num = num * Par.y_unit;             Cp++;   break;
        default:                                                break;
    }
    n = (int) (num + 0.5);
    *val = n * s;
#endif
    return 1;
}

int isfnchar(char c)
{
    if (isalnum(c)
    ||  c == '/'
    ||  c == '.'
    ||  c == '_')
        return 1;
    return 0;
}

int getfname(char *buf, int siz)
{
    int     i = 0;

    while (isfnchar(*Cp))
    {
        if (i >= siz)
            return 0;
        buf[i++] = *Cp++;
    }
    buf[i] = 0;
    if (i > 0)
        return 1;
    return 0; 
}

char *push_str(char *str)
{
    int     len;
    char    *p;

    len = strlen(str) + 1;
    p = calloc(1, len);
    if (p == NULL)
        return NULL;
    memcpy(p, str, len);
    return p;
}

